#!/usr/bin/env python3

from optparse import OptionParser
import os
import io
import sys
import json
import copy
import logging

from collections import defaultdict, deque, OrderedDict

from confluent_kafka import Producer
from cronut import App
from cronut.utils import uriparse

from ligo.lw import ligolw
from ligo.lw import lsctables
from ligo.lw import utils as ligolw_utils

from lal import GPSTimeNow

from ligo.scald.io import kafka

from gw.lts import utils

def parse_command_line():
	parser = utils.add_general_opts()
	parser.add_option('--preferred-event', metavar = 'func:param', default = 'max:snr', help = 'Parameter and function to use to determine preferred events in the case that multiple event messages are found for a single injection. ' +
		'Supported options are min:combined_far, max:snr, latest:msg_time, or first:msg_time')
	opts, args = parser.parse_args()

	return opts, args

class InspInjMsgFind(object):
	def __init__(self, options):
		self.tag = options.tag
		self.kafka_server = options.kafka_server
		self.topics = options.input_topic
		self.preferred_func = utils.preferred_event_func(options.preferred_event.split(':')[0])
		self.preferred_param = options.preferred_event.split(':')[1]
		self.verbose = options.verbose

		# initialize data deques
		# if injections come every ~20 seconds this should correspond 
		# to keeping messages for about 3-4 minutes.
		self.maxlen = 10
		self.event_msgs = defaultdict(lambda: deque(maxlen=self.maxlen))
		self.inj_msgs = defaultdict(lambda: deque(maxlen=self.maxlen))
		self.msgs_sent = defaultdict(lambda: OrderedDict())

		# set up producer
		self.client = kafka.Client(f'kafka://{self.tag}@{self.kafka_server}')

		# create a job service using cronut
		self.app = App('inspinjmsg_find', broker=f'kafka://{self.tag}_inspinjmsg_find@{self.kafka_server}')

		@self.app.process(self.topics)
		def process(message):
			mtopic = message.topic().split('.')[-1]
			mpipeline = message.topic().split('.')[0]

			# unpack data from the message
			if mtopic == 'inj_events':
				# parse event info
				event = json.loads(message.value())

				# load the coinc table
				coinc_file = utils.load_xml(event['coinc'])
				coinctable = lsctables.CoincInspiralTable.get_table(coinc_file)

				# get event coalescence time
				coinctime = coinctable[0].end_time + 10.**-9. * coinctable[0].end_time_ns
				logging.info(f'received {mpipeline} event with coalescence time: {coinctime}')

				dict = {
					'time': coinctime,
					'coinc': coinc_file,
					'msg_time': int(GPSTimeNow()),
				}

				# add optional keys - these may or may not
				# already be present depending on the data
				# source configuration
				for key in ['latency', 'p_astro', 'uid']:
					if key in event.keys():
						dict.update({key: event[key]})
					else:
						dict.update({key: None})

				# store event data
				self.event_msgs[mpipeline].append(dict)

				self.process_events(mpipeline)
				self.process_stale_msgs(mtopic, mpipeline)

			elif mtopic == 'inj_stream':
				# parse inj info
				injection = json.loads(message.value())
				ifos = injection['onIFOs']

				# load the sim table
				simfile = utils.load_xml(injection['sim'])
				simtable = lsctables.SimInspiralTable.get_table(simfile)

				# get injection coalescence time
				coinctime = simtable[0].geocent_end_time + 10.**-9 * simtable[0].geocent_end_time_ns
				logging.info(f'received {mpipeline} injection with coalescence time: {coinctime}')

				# store inj data
				self.inj_msgs[mpipeline].append({
								'time': coinctime,
								'sim': simfile,
								'ifos': ifos
				})

				self.process_events(mpipeline)
				self.process_stale_msgs(mtopic, mpipeline)

			else:
				# break
				logging.debug(f'Error: Found unexpected message from topic {mtopic}.')


	def start(self):
		# start up
		logging.info('Starting up...')
		self.app.start()


	def append_sim_table(self, coinc_file, sim_file):
		# init a new sim inspiral table
		this_sim_table = lsctables.SimInspiralTable.get_table(sim_file)
		coinc_file.childNodes[-1].appendChild(this_sim_table)

		return coinc_file


	def write_sim_file(self, sim, xmldoc):
		# open a new xml doc
		sim_msg = io.BytesIO()
		ligolw_elem = xmldoc.appendChild(ligolw.LIGO_LW())

		output_simtable = ligolw_elem.appendChild(lsctables.New(lsctables.SimInspiralTable))
		this_sim_table = lsctables.SimInspiralTable.get_table(sim)
		output_simtable.extend(this_sim_table)
		ligolw_utils.write_fileobj(xmldoc, sim_msg, gz = False)

		return sim_msg


	def construct_event_ouput(self, xmldoc, event, injection, key=None):
		filename = f'coinc-{int(event["time"])}.xml' if not key else f'{key}-coinc-{int(event["time"])}.xml'

		coinctable = lsctables.CoincInspiralTable.get_table(event['coinc'])

		ligolw_utils.write_filename(xmldoc, os.path.join('coincs', filename), verbose = self.verbose)
		coinc_msg = io.BytesIO()
		ligolw_utils.write_fileobj(xmldoc, coinc_msg, gz = False)

		output = {
			'time': coinctable[0].end_time,
			'time_ns': coinctable[0].end_time_ns,
			'snr': coinctable[0].snr,
			'far': coinctable[0].combined_far,
			'p_astro': event['p_astro'],
			'coinc': coinc_msg.getvalue().decode(),
			'latency': event['latency'],
			'uid': event['uid'],
			'onIFOs': injection['ifos']
			}

		return output


	def process_events(self, pipeline):
		# for each event in the event_msgs deque, find the nearest injection in inj_msgs
		# within +/- delta_t (1 second) of the event coalescence time.
		# when an association is made, remove both messages from the deques,
		# add the sim inspiral table from injection to the event's coinc xml and 
		# send a message to the testsuite.events topic for all the other jobs to consume
		events_copy = copy.copy(self.event_msgs[pipeline])
		injections = copy.deepcopy(self.inj_msgs[pipeline])
		for event in events_copy:
			event_time = event['time']

			nearest_inj = utils.find_nearest_msg(injections, event_time)

			# if no associated injection was found, pass
			if not nearest_inj:
				continue

			inj_time = nearest_inj['time']
			sim_file = nearest_inj['sim']

			# remove the event from the deque and unpack info
			self.event_msgs[pipeline].remove(event)
			coinc_file = event['coinc']
			this_coinc = lsctables.CoincInspiralTable.get_table(coinc_file)
			this_sngl = lsctables.SnglInspiralTable.get_table(coinc_file)
			this_coinc_event = lsctables.CoincTable.get_table(coinc_file)

			time_now = int(GPSTimeNow())
			if self.preferred_param == 'combined_far' or self.preferred_param == 'snr':
				val = this_coinc.getColumnByName(self.preferred_param)[0]
			elif self.preferred_param == 'msg_time':
				val = event['msg_time']
			else:
				raise NotImplementedError

			# Note: this requires that aggregate by "latest" works the way we would hope
			if inj_time in self.msgs_sent[pipeline].keys():
				vals = list(self.msgs_sent[pipeline][inj_time].values()) + [val]
				if not val is self.preferred_func(vals):
					logging.debug(f'New event for injection: {inj_time}. This {self.preferred_param} {val} not preferred to send an update msg, continuing.')
					continue
				# wait some time before sending an update message
				if not time_now - max(self.msgs_sent[pipeline][inj_time].keys()) >= 10.:
					continue
				logging.debug(f'New event for injection {inj_time}. This {self.preferred_param} {val} is preferred, sending an update msg')
				self.msgs_sent[pipeline][inj_time].update({time_now: val})
			else:
				logging.debug(f'Initial event found for injection {inj_time} with {self.preferred_param} {val}.')
				self.msgs_sent[pipeline][inj_time] = {time_now: val}

			# store relevant info for missed/found table in coinc_dict_list
			coinc_dict = {}

			for attr in ("combined_far", "snr", "false_alarm_rate"):
				try:
					coinc_dict[attr] = float(this_coinc.getColumnByName(attr)[0])
				except TypeError:
					pass
			coinc_dict["end"] = float(event_time)

			for attr in ("likelihood",):
				try:
					coinc_dict[attr] = this_coinc_event.getColumnByName(attr)[0]
				except TypeError:
					pass

			for sngl_row in this_sngl:
				for attr in ("snr", "chisq", "mass1", "mass2", "spin1z", "spin2z", "coa_phase"):
					if getattr(sngl_row, attr):
						coinc_dict["%s_%s" % (sngl_row.ifo, attr)] = float(getattr(sngl_row, attr))
				if sngl_row.end:
					coinc_dict["%s_end" % sngl_row.ifo] = float(sngl_row.end)

			# proceed with sending event
			# add sim table to coinc file and write to disk
			newxmldoc = self.append_sim_table(coinc_file, sim_file)

			output = self.construct_event_ouput(newxmldoc, event, nearest_inj)

			self.client.write(f'{pipeline}.{self.tag}.testsuite.events', output)
			logging.info(f'Sent msg to: {pipeline}.{self.tag}.testsuite.events')
			newxmldoc.unlink()

			# send inj_table_events message
			self.client.write(f'{pipeline}.{self.tag}.testsuite.inj_table_events', [coinc_dict])
			logging.info(f'Sent msg to: {pipeline}.{self.tag}.testsuite.inj_table_events')


	def process_stale_msgs(self, topic, pipeline):
		# process old messages: either messages that are about to be 
		# removed from the left of the deque, or have been in the deque
		# for 2 hours, and send a message with the necessary info
		# this is necessary in the case that:
			# 1) we receive an event from the search which is not
			# associated with an injection, ie a glitch or real gw 
			# candidate.
			# 2) there is an injection for which we never receive
			# an associated event from the search. ie the injection
			# was not recovered at even the GDB far threshold.
		# FIXME dont hardcode wait time
		if self.inj_msgs[pipeline] and ((len(self.inj_msgs[pipeline]) == self.maxlen) or (float(GPSTimeNow()) - self.inj_msgs[pipeline][0]['time'] >= 7200.)):
			stale_msg = self.inj_msgs[pipeline][0]

			if topic == 'inj_events':
				logging.debug(f'{pipeline} event from time {stale_msg["time"]} to be removed from the queue - no associated injection found')

			elif topic == 'inj_stream':
				sim_inspiral = stale_msg['sim']

				if not stale_msg['time'] in self.msgs_sent[pipeline].keys():
					logging.debug(f'Sending {pipeline} missed injection msg for injection {stale_msg["time"]}')
					newxmldoc = ligolw.Document()
					sim_msg = self.write_sim_file(sim_inspiral, newxmldoc)

					output = {
							'sim': sim_msg.getvalue().decode(),
							'onIFOs': stale_msg['ifos'],
					}

					self.client.write(f'{pipeline}.{self.tag}.testsuite.missed_inj', output)
					logging.info(f'Sent msg to: {pipeline}.{self.tag}.testsuite.missed_inj')
					newxmldoc.unlink()


def main():
	# parse options from command line
	opts, args = parse_command_line()

	# set up logging
	utils.set_up_logger(opts.verbose)

	# set up dir for output coincs
	try:
		os.mkdir('coincs')
	except OSError as error:
		pass

	# initialize the processor
	processor = InspInjMsgFind(opts)
	processor.start()

if __name__ == '__main__':
	main()



#!/usr/bin/env python3

from optparse import OptionParser
import json
import logging
import copy

from collections import defaultdict, deque

from confluent_kafka import Producer
from cronut import App
from cronut.utils import uriparse

from gw.lts import utils
from gw.lts.utils.gracedb_helper import GraceDbHelper

from ligo.scald.io import kafka

from ligo.lw import lsctables

def parse_command_line():
	parser = utils.add_general_opts()
	parser.add_option('--pastro-file', metavar = 'name=file', action = 'append', help = 'Name of the p(astro) method (ex. fgmc or mchirp) and filename to download from GraceDB, given as name=file. Can be given multiple times.')
	parser.add_option('--gdb-pastros', action = 'store_true', default = False, help = 'Download p(astro) files from GraceDB. This will fail if the data source is fake-data.')
	opts, args = parser.parse_args()

	return opts, args

class PAstro(object):
	def __init__(self, options):
		self.tag = options.tag
		self.kafka_server = options.kafka_server
		self.gdb_pastros = options.gdb_pastros

		self.pastro_files = {}
		for option in options.pastro_file:
			name, file = option.split('=')
			self.pastro_files.update({name: file})

		# set up producer
		self.client = kafka.Client(f'kafka://{self.tag}@{options.kafka_server}')

		if self.gdb_pastros:
			self.gracedb_helper = GraceDbHelper(options.group)

		# initialize output dict
		self.events = {name: deque(maxlen=10) for name in self.pastro_files.keys()}

		# create a job service using cronut
		self.app = App('pastro', broker=f'kafka://{self.tag}_pastro@{self.kafka_server}')

		# subscribes to a topic
		@self.app.process(options.input_topic)
		def process(message):
			mtopic = message.topic().split('.')[-1]
			mpipeline = message.topic().split('.')[0]
			mkey = utils.parse_msg_key(message)
			logging.info(f'Read message from {mpipeline} {mtopic}.')

			# parse message value
			event = json.loads(message.value())
			event.update({'pipeline': mpipeline})

			for name, file in self.pastro_files.items():
				response = self.process_event(event, name, file)
				if not response:
					# keep track of events that failed
					# to get a p(astro) on the first try
					# when getting p(astro)s from gracedb, this
					# can happen if the p(astro) isnt uploaded
					# immediately
					times = [e["time"] for e in self.events[name]]
					if not event["time"] in times:
						self.events[name].append(event)

			# iterate over events and try again to grab a
			# p(astro) for each one. On success, remove the
			# event from the deque
			for name, file in self.pastro_files.items():
				for e in copy.deepcopy(self.events[name]):
					response = self.process_event(e, name, file)
					if response:
						self.events[name].remove(e)

	def start(self):
		# start up
		logging.info('Starting up...')
		self.app.start()


	def process_event(self, event, pastro_name, filename):
		if self.gdb_pastros:
			file = self.gracedb_helper.query_file(event['uid'], filename = filename)
			if file:
				p_astro_dict = json.loads(file.read())
			else:
				p_astro_dict = None

		else:
			try:
				p_astro_dict = json.loads(event['p_astro'])
			except KeyError:
				raise KeyError("Event message does not include p(astro) and --gdb-pastros is not given. There is no way to retrieve the p(astro) in this case.")

		if p_astro_dict:
			output = {}
			time = event["time"]
			pipeline = event["pipeline"]

			p_astro_dict['astro'] = 1 - p_astro_dict['Terrestrial']

			# determine source from inspiral table
			coinc_file = utils.load_xml(event['coinc'])
			simtable = lsctables.SimInspiralTable.get_table(coinc_file)
			source = utils.source_tag(simtable)

			for key, value in p_astro_dict.items():
				output['p_' + key] = {
					'time': [ float(time) ],
					'data': [ float(value) ]
				}

			logging.debug(f'{source} event: {pastro_name}: {p_astro_dict}')

			# send message to output topics
			for topic, data in output.items():
				self.client.write(f'{pipeline}.{self.tag}.testsuite.{topic}', data, tags = [pastro_name, source])
				logging.info(f'Sent output message to output topic: {pipeline}.{self.tag}.testsuite.{topic}.')

			return True

		else:
			return False


def main():
	opts, args = parse_command_line()

	# set up logging
	utils.set_up_logger(opts.verbose)

	processor = PAstro(opts)
	processor.start()

if __name__ == '__main__':
	main()



#!/usr/bin/env python

import os
import http.client

from ligo.gracedb.rest import GraceDb
from ligo.gracedb.exceptions import HTTPError

class GraceDbHelper(object):
	def __init__(self, group='gracedb-playground'):
		self.group = group
		self.client = GraceDb(service_url=f'https://{self.group}.ligo.org/api/')

	def get_event(self, uid=None, time=None, pipeline=None, search=None):
		try:
			if uid:
				return self.client.event(uid)

			query = f'{time - 1.} .. {time + 1.} '
			if pipline:
				query += f'pipeline: {pipeline} '

			if search:
				query += f'search: {search} '

			event = next(self.client.events(query))

			return event
		except HTTPError:
			return None
		
	def query_file(self, graceid, filename, outpath=None, tag=None):
		"""
		Download a file from given GraceDb event.
		Optionally write out the file to disk.
		Return the file.
		"""
		try:
			response = self.client.files(graceid, filename)
			if response.status == http.client.OK:
				if outpath:
					outfile = os.path.join(outpath, f'{tag}-{graceid}.fits')
					with open(outfile, 'wb') as f:
						f.write(response.read())
		except HTTPError:
			response = None
	
		return response



#!/usr/bin/env python3

import json
import logging
import io

from igwn_alert.client import client as IGWNAlertClient

from ligo.lw import ligolw
from ligo.lw import lsctables
from ligo.lw import utils as ligolw_utils

from optparse import OptionParser
from confluent_kafka import Producer

from collections import OrderedDict, deque

from ligo.scald.io import kafka

from lal import GPSTimeNow

from gw.lts import utils
from gw.lts.utils.gracedb_helper import GraceDbHelper

class LIGOLWContentHandler(ligolw.LIGOLWContentHandler):
	pass

lsctables.use_in(LIGOLWContentHandler)

def parse_command_line():
	parser = utils.add_general_opts()
	parser.add_option('--gdb-topics', metavar='string', action = 'append', help = 'GraceDb topics to subscribe to. Can be given multiple times.')
	parser.add_option('--max-wait-time', metavar = 'float', default = 3600., help = 'Max amount of time to keep events before removing them, whether or not a message has been sent')
	opts, args = parser.parse_args()
	return opts, args

def main():
	# parse command line
	opts, args = parse_command_line()
	
	## set up logging
	utils.set_up_logger(opts.verbose)

	# set up listener
	listener = on_alert(opts)

	# initialize a client and listener
	client = IGWNAlertClient(group=opts.group)

	client.listen(listener.process_alert, opts.gdb_topics)

class on_alert(object):
	def __init__(self, options):
		self.tag = options.tag
		self.kafka_server = options.kafka_server
		self.gracedb_client = GraceDbHelper(options.group)
		self.max_wait_time = options.max_wait_time
		self.inj_channels = set(['GDS-CALIB_STRAIN_INJ1_O3Replay', 'GDS-CALIB_STRAIN_INJ1_O3Replay', 'Hrec_hoft_16384Hz_INJ1_O3Replay'])

		# set up producer
		self.client = kafka.Client(f'kafka://{self.tag}@{self.kafka_server}')

		self.events = OrderedDict()
		self.events_sent = deque(maxlen=300)

		logging.info('Initialized on_alert class.')

	def process_alert(self, topic=None, payload=None):
		# unpack data
		uid = payload['uid']
		alert_type = payload['alert_type']
		data = payload['data']
		object = payload['object']

		channels = self.get_channels(uid)
		if not channels or not channels.issubset(self.inj_channels):
			logging.debug(f'{uid} not from injection channels, skipping.')
			return

		# find the pipeline that uploaded this event
		# this is encoded in the output topic when the message is sent
		pipeline = topic.split("_")[1]

		if alert_type == 'log' or alert_type == 'new':
			logging.info(f'Received {alert_type} alert for {uid} from {pipeline}')

		if uid in self.events.keys():
			self.events[uid] = self.process_event(uid, output=self.events[uid])
		else:
			self.events[uid] = self.process_event(uid)

		# check if all elements present, then send msg
		# only send msg once per event
		if all(self.events[uid].values()) and not uid in self.events_sent:
			logging.info(f'sending a message for {uid} (coa time: {self.events[uid]["time"]})...')
			self.client.write(f'{pipeline}.{self.tag}.inj_events', self.events[uid])
			logging.info(f'Sent msg to: {pipeline}.{self.tag}.inj_events')
			self.events_sent.append(uid)

		# remove old msgs that already had a msg sent
		time_now = float(GPSTimeNow())
		for key, value in list(self.events.items()):
			if time_now - value['time_added'] >= self.max_wait_time:
				logging.debug(f'Removing old event: {key}')
				self.events.pop(key)

	def process_event(self, uid, output={}):
		if not output:
			# initialize all the items we need in order to send a message
			for k in ['time', 'time_ns', 'snr', 'far', 'coinc', 'latency']:
				output.update({k: None})
			output.update({'time_added': float(GPSTimeNow())})

		output.update({'uid': uid})
		output.update(self.add_coinc(uid))
		output.update(self.add_latency(uid))

		return output

	def add_coinc(self, uid, output={}):
		coinc = self.get_filename(uid, 'coinc.xml')
		if coinc:
			xmldoc = utils.load_xml(coinc)
			coinctable = lsctables.CoincInspiralTable.get_table(xmldoc)

			coinc_msg = io.BytesIO()
			ligolw_utils.write_fileobj(xmldoc, coinc_msg, gz = False)

			output.update({
					'time': coinctable[0].end_time,
					'time_ns': coinctable[0].end_time_ns,
					'snr': coinctable[0].snr,
					'far': coinctable[0].combined_far,
					'coinc': coinc_msg.getvalue().decode()
			})	
			logging.debug(f'Added coinc.xml to {uid}')
		return output

	def add_latency(self, uid, output = {}, retries=10):
		this_try = 0
		while this_try < retries:
			event = self.gracedb_client.get_event(uid=uid)
			if event:
				event = event.json()
				output.update({
						'latency': event['reporting_latency']
				})
				logging.debug(f'Added latency: {event["reporting_latency"]}')
				return output
			this_try += 1
		logging.debug(f'Failed to retrieve {uid} latency.')
		return None

	def get_channels(self, uid):
		coinc = self.get_filename(uid, 'coinc.xml')
		if coinc:
			xmldoc = utils.load_xml(coinc)
			sngltable = lsctables.SnglInspiralTable.get_table(xmldoc)
			channels = set(list(sngltable.getColumnByName('channel')))

			return channels
		return None

	def get_filename(self, uid, filename, retries=10):
		this_try = 0
		while this_try < retries:
			file = self.gracedb_client.query_file(uid, filename)
			if file:
				return file
			else:
				this_try += 1
		logging.debug(f'Failed to download {filename} from {uid}.')
		return None

if __name__ == '__main__':
	main()



#!/usr/bin/env python3

from optparse import OptionParser
import os
import sys
import json
import logging
import math
import atexit

from collections import defaultdict, deque

from confluent_kafka import Producer
from cronut import App
from cronut.utils import uriparse

from ligo.lw import lsctables
from ligo.lw import utils as ligolw_utils
from ligo.lw.utils.process import register_to_xmldoc

from lal import GPSTimeNow

from ligo.scald.io import kafka

from gw.lts import utils
from gw.lts.utils import cosmology_utils as cutils

def parse_command_line():
	parser = utils.add_general_opts()
	parser.add_option('--far-threshold', default=2.314e-5,  help = 'far threshold for missed vs found injections. Default is 2 per day.')
	parser.add_option('--calculate-injected-vt', action='store_true', help='If the injection file doesnt already have an injected VT calculated and stored in Process Params Table , calculate it here on the fly.' + 
			'In this case, max_redshift is required.')
	parser.add_option('--max-redshift', metavar = 'SOURCE:float', action = 'append', help='The max redshift used when generating the injection set.' +
			'Required if --calculate-injected-vt is set. Can be given multiple times.')
	parser.add_option('--bootstrap-vt', action='store_true', default=False, help = 'Whether to load counts for previous found injections from the injection set.' + 
			'This is used to calculate a cumulative VT even if the job is re-started.')
	opts, args = parser.parse_args()

	return opts, args

def write_counts_to_disk(num_found, exp_num_found, filename):
	xmldoc = utils.load_filename(filename)
	process_params_table = lsctables.ProcessParamsTable.get_table(xmldoc)

	process_params_table, new_vt_rows = process_update(num_found, process_params_table, rowname = 'cumulative-vt')
	process_params_table, new_exp_vt_rows = process_update(exp_num_found, process_params_table, rowname = 'cumulative-exp-vt')

	new_vt_rows.update(new_exp_vt_rows)
	if new_vt_rows:
		register_to_xmldoc(xmldoc, program = 'vt', paramdict = new_vt_rows, comment = 'Cumulative VT')

	ligolw_utils.write_filename(xmldoc, filename, verbose = True)

def process_update(dict, table, rowname):
	new_rows = {}
	for pipeline in dict.keys():
		for key, data in dict[pipeline].items():
			logging.debug(f'At shutdown: {pipeline} {key} {rowname} last time: {data[-1][0]} | last count: {data[-1][1]}')
			updated = 0
			# if the row already exists, update its value
			for r in table:
				if str(r.param).strip('--') == f'{rowname}:{pipeline}-{key}':
					logging.debug(f'found row to update: {str(r.param)}...')
					r.value = f'{data[-1][0]}:{data[-1][1]}'
					updated += 1
			# otherwise, add a new row
			if not updated:
				logging.debug('no row to update, writing a new row...')
				new_rows[f'{rowname}:{pipeline}-{key}'] =  f'{data[-1][0]}:{data[-1][1]}'

	return table, new_rows

def parse_vt_from_process(row):
	params = str(row.param).split(':')[1]
	pipeline = '-'.join(params.split('-')[:-1])
	source = params.split('-')[-1]
	time, count = str(row.value).split(':')

	return pipeline, source, time, count

def injected_VT(pptable, options):
	# parse process params table
	ppdict = {}
	for r in pptable:
		try:
			ppdict.setdefault(str(r.param).strip('--'), []).append(float(r.value))
		except:
			ppdict.setdefault(str(r.param).strip('--'), []).append(r.value)

	# First get total number of injections
	if 'accept' in ppdict and 'reject' in ppdict:
		num_total_injections = sum(ppdict['accept']) + sum(ppdict['reject'])
	elif 'total-generated' in ppdict:
		num_total_injections = sum(ppdict['total-generated'])
	else:
		raise Exception('Could not find total number of attempted injections in file, exiting')

	# try to calculate VT myself
	if num_total_injections and options.calculate_injected_vt and options.max_redshift:
		try:
			omega = cutils.get_cosmo_params()
			gps_start = ppdict['gps-start'][0]
			gps_end = ppdict['gps-end'][0]

			VT = {}
			for z in options.max_redshift:
				source = z.split(':')[0].upper()
				val = float(z.split(':')[1])
			
				VT.update({source: cutils.surveyed_spacetime_volume(gps_start, gps_end, val, omega)})
				logging.debug(f'{source} VT: {VT[source]}')
			VT.update({'VT': sum(VT[source] for source in VT.keys())})

			return VT, num_total_injections
		except:
			 raise Exception('Could not calculate the injected VT, exiting')

	# try getting VT from the injection file
	else:
		try:
			VT = {}
			for item in ['bns-vt', 'nsbh-vt', 'bbh-vt', 'VT']:
				if item in ppdict:
					key = item.split('-')[0].upper()
					VT.setdefault(key, sum(ppdict[item]))
					logging.debug(f'{key} VT: {sum(ppdict[item])}')

			return VT, num_total_injections

		except:
			 raise Exception('Could not get VT from the Process Params Table, exiting')

def main():
	opts, args = parse_command_line()
	
	tag = opts.tag
	far_threshold = float(opts.far_threshold)
	pipelines = [x.split('.')[0] for x in opts.input_topic]
	
	# set up producer
	client = kafka.Client(f'kafka://{tag}@{opts.kafka_server}')
	
	# set up logging
	utils.set_up_logger(opts.verbose)
	
	# load and parse injection file for injected VT and total injections
	inj_file = utils.load_filename(opts.inj_file, gz=True)
	ProcessParamsTable = lsctables.ProcessParamsTable.get_table(inj_file)
	
	inj_VT, total_inj = injected_VT(ProcessParamsTable, opts)
	
	if not inj_VT:
		raise Exception('Injected VT dict is empty, exiting')
	
	# initialize dicts to store data
	num_found = defaultdict(lambda: defaultdict(lambda: deque(maxlen=300)))
	expected_num_found = defaultdict(lambda: defaultdict(lambda: deque(maxlen=300)))
	output = defaultdict(lambda: {'time': [], 'data': []})
	
	# init num found to 0 for every ifo combination
	startup_time = float(GPSTimeNow())
	
	# add previous found injection counts first
	if opts.bootstrap_vt:
		for r in ProcessParamsTable:
			if 'cumulative-vt' in str(r.param):
				pipeline, source, time, count = parse_vt_from_process(r)
				num_found[pipeline][source].append((float(time), float(count)))
				logging.debug(f'Bootstrapping found injections for {pipeline} {source} from time: {time}, count: {count}')
	
			if 'cumulative-exp-vt' in str(r.param):
				pipeline, source, time, count = parse_vt_from_process(r)
				expected_num_found[pipeline][source].append((float(time), float(count)))
				logging.debug(f'Bootstrapping expected found injections for {pipeline} {source} from time: {time}, count: {count}')
	
	# for any pipelines or sources that didn't have a previously recorded value
	# in the injection file, start the count from zero
	for pipeline in pipelines:
		for key in inj_VT:
			if not num_found[pipeline][key]:
				if opts.bootstrap_vt:
					logging.debug(f'No {pipeline} {key} counts found to bootstrap. Starting count from 0.')
				num_found[pipeline][key].append((startup_time, 0))
			if not expected_num_found[pipeline][key]:
				if opts.bootstrap_vt:
					logging.debug(f'No {pipeline} {key} expected counts found to bootstrap. Starting count from 0.')
				expected_num_found[pipeline][key].append((startup_time, 0))
	
	# create a job service using cronut
	app = App('vt', broker=f'kafka://{tag}_vt@{opts.kafka_server}')
	
	# subscribes to a topic
	@app.process(opts.input_topic)
	def process(message): 
		mtopic = message.topic().split('.')[-1]
		mpipeline = message.topic().split('.')[0]
		mkey = utils.parse_msg_key(message)
		logging.debug(f'Read message from {mpipeline} {mtopic}')
	
		# dont process old messages or noninjection events
		event = json.loads(message.value())
		time = event['time'] + event['time_ns'] * 10**-9.
		if time > startup_time and not mkey == 'noninj':
			# parse event info
			coinc_file = utils.load_xml(event['coinc'])
			simtable = lsctables.SimInspiralTable.get_table(coinc_file)
	
			far = event['far']
			snr = event['snr']
	
			source = utils.source_tag(simtable)
	
			# try calculating source specific VTs, otherwise just calculalte overall VT
			if source in inj_VT:
				key = source
			else:
				key = 'VT'
	
			# calculate decisive SNR and count expected number of found injections
			ifos = event['onIFOs']
			inj_snrs = defaultdict(lambda: None)
			inj_snrs['H1'] = simtable[0].alpha4
			inj_snrs['L1'] = simtable[0].alpha5
			inj_snrs['V1'] = simtable[0].alpha6
	
			decisive_snr = utils.decisive_snr(inj_snrs, ifos)
	
			if decisive_snr:
				prev_exp_found = expected_num_found[mpipeline][key][-1][1]
				this_expected_found = prev_exp_found + 1 if decisive_snr >= 8. else prev_exp_found
				expected_num_found[mpipeline][key].append((time, this_expected_found))
	
			# Count the actual  number of found injections so far
			prev_num_found = num_found[mpipeline][key][-1][1]
			this_num_found = prev_num_found + 1 if far < far_threshold else prev_num_found
			num_found[mpipeline][key].append((time, this_num_found))
	
			# Calculate the current and expected VT
			if decisive_snr:
				expected_VT = (this_expected_found / total_inj) * inj_VT[key]
				output['exp_vt'] = {
					'time': [ time ],
					'data': [ expected_VT ]
				}
			VT = (this_num_found / total_inj) * inj_VT[key]
	
			output['vt'] = {
				'time': [ time ],
				'data': [ VT ]
			}
	
			# Calculate the current and expected sensitive volume (scaled for time)
			# divide the VT by the total elapsed time
			dt = abs(time - startup_time) / 60. / 60. / 24. / 365.25 # yr
			if decisive_snr:
				expected_V = expected_VT / dt
				output['exp_sensitive_vol'] = {
					'time': [ time ],
					'data': [ expected_V ]
				}
			V = VT / dt
	
			output['sensitive_vol'] = {
				'time': [ time ],
				'data': [ V ]
			}
	
	
			# Calculate the current and expected range, assuming V is a sphere
			if decisive_snr:
				expected_R = (expected_V / (4. * math.pi / 3.))**(1./3.)
				output['exp_range'] = {
					'time': [ time ],
					'data': [ expected_R ]
				}
			R = (V / (4. * math.pi / 3.))**(1./3.)
	
			output['range'] = {
				'time': [ time ], 
				'data': [ R ]
			}
	
			if decisive_snr:
				logging.debug(f'{mpipeline}: {source} expec. VT: {expected_VT} Gpc3 yr | expc. V: {expected_V} Gpc3 | expec. range: {expected_R} Gpc')
			logging.info(f'{mpipeline}: {source} current VT: {VT} Gpc3 yr | V: {V} Gpc3 | range: {R} Gpc')
			for topic in output:
				client.write(f'{mpipeline}.{tag}.testsuite.{topic}', output[topic], tags = source)
				logging.info(f'Sent msg to: {mpipeline}.{tag}.testsuite.{topic}')

	# at exit, writ counts to disk
	atexit.register(write_counts_to_disk, num_found, expected_num_found, opts.inj_file)
	
	# start up
	logging.info('Starting up...')
	app.start()

if __name__ == '__main__':
	main()



#  Copyright (C) 2017 Jolien Creighton
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with with program; see the file COPYING. If not, write to the
#  Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#  MA  02111-1307  USA

## @file
# The python module for utilities needed for cosmological calculations.

#
# =============================================================================
#
#                                   Preamble
#
# =============================================================================
#

import numpy
import scipy.integrate

import lal


#
# =============================================================================
#
#                      Cosmological and Other Constants
#
# =============================================================================
#

# FIXME: go from astropy instead
def get_cosmo_params():
    # From Planck2015, Table IV
    omega = lal.CreateCosmologicalParametersAndRate().omega
    lal.SetCosmologicalParametersDefaultValue(omega)
    omega.h = 0.679
    omega.om = 0.3065
    omega.ol = 0.6935
    omega.ok = 1.0 - omega.om - omega.ol
    omega.w0 = -1.0
    omega.w1 = 0.0
    omega.w2 = 0.0

    return omega

#
# =============================================================================
#
#                               Cosmology Utilities
#
# =============================================================================
#


def surveyed_spacetime_volume(gps_start_time, gps_end_time, max_redshift, omega):
    '''
    Returns the total spacetime volume surveyed:

        <VT> = T \int dz \frac{dV_c}{dz} \frac{1}{1+z}

    Results are given in units Gpc^3 yr(Julian).
    '''

    # Note: LAL's cosmology routines returns distances in Mpc
    def integrand(z, omega):
        '''
        Returns the integrand

            (1 + z) D_A^2(z) / E(z)

        in units of Mpc^2.  Multiply the integral by 4 * pi * D_H
        to get the desired integral.
        '''

        return (1.0 + z) * lal.AngularDistance(omega, z)**2 * lal.HubbleParameter(z, omega)

    I, _ = scipy.integrate.quad(integrand, 0.0, max_redshift, args=omega)

    # multiply by remaining factors and scale to Gpc^3
    V = 4.0 * numpy.pi * lal.HubbleDistance(omega) * I / (1e3)**3
    
    # surveyed time in Julian years
    T = (gps_end_time - gps_start_time) / lal.YRJUL_SI

    return V * T





#!/usr/bin/env python3

from optparse import OptionParser
import os
import sys
import json
import logging

from collections import defaultdict, deque

from confluent_kafka import Producer
from cronut import App
from cronut.utils import uriparse

from ligo.lw import lsctables

from lal import GPSTimeNow

from ligo.scald.io import kafka

from gw.lts import utils

def parse_command_line():

	parser = utils.add_general_opts()
	opts, args = parser.parse_args()

	return opts, args

def main():
	opts, args = parse_command_line()
	
	tag = opts.tag
	
	# set up producer
	client = kafka.Client(f'kafka://{tag}@{opts.kafka_server}')
	
	# set up logging
	utils.set_up_logger(opts.verbose)
	
	# initialize data deque and output dict
	output = defaultdict(lambda: {'time': [], 'data': []})
	data = deque(maxlen=300)
	
	# create a job service using cronut
	app = App('likelihood', broker=f'kafka://{tag}_likelihood@{opts.kafka_server}')
	
	# subscribes to a topic
	@app.process([f'{opts.data_source}.{tag}.testsuite.{topic}' for topic in opts.input_topic])
	def process(message): 
		mtopic = message.topic()
		mkey = utils.parse_msg_key(message)
		logging.debug(f'Read message from {mtopic}.')
	
		# dont process noninjection events
		if not mkey == 'noninj':
			# parse event info
			event = json.loads(message.value())
	
			coinc_file = utils.load_xml(event['coinc'])
			far = event['far']
			snr = event['snr']
			time = event['time'] + event['time_ns'] * 10**-9.
	
			# load likelihood from coinc event table
			event_table = lsctables.CoincTable.get_table(coinc_file)
			likelihood = event_table[0].likelihood
	
			# output message
			output['likelihood']['time'].append(float(time))
			output['likelihood']['data'].append(float(likelihood))
	
			# send message to output topics
			for key in output:
				client.write(f'{opts.data_source}.{tag}.testsuite.{key}', output[key])
				logging.info(f'Sent output message to output topic: {opts.data_source}.{tag}.testsuite.{key}.')

	# start up
	logging.info('Starting up...')
	app.start()

if __name__ == '__main__':
	main()



#!/usr/bin/env python3

from optparse import OptionParser
import os
import sys
import json
import logging

from collections import defaultdict, deque

from confluent_kafka import Producer
from cronut import App
from cronut.utils import uriparse

from ligo.lw import lsctables

from lal import GPSTimeNow

from ligo.scald.io import kafka

from gw.lts import utils

def parse_command_line():
	parser = utils.add_general_opts()

	parser.add_option('--far-threshold', default=2.314e-5, help = 'far threshold for missed vs found injections. Default is 2 per day.')
	opts, args = parser.parse_args()

	return opts, args

def construct_output(injected_snrs = {}, recovered_snrs = {}, time = None, decisive_snr = None):
	output = defaultdict(lambda: {'time': [], 'data': []})

	for ifo, value in injected_snrs.items():
		if not value:
			continue
		output[f'{ifo}_injsnr'] = {
			'time': [ time ],
			'data': [ value ]
		}

	for ifo, value in recovered_snrs.items():
		if not value:
			continue
		output[f'{ifo}_recsnr'] = {
			'time': [ time ],
			'data': [ value ]
		}

	output['decisive_snr'] = {
			'time': [ time ],
			'data': [decisive_snr]
	}

	return output

def process_event(coinc_file, on_ifos, key):
		snrs = defaultdict(lambda: 0)

		# load coinc file and get data
		coinctable = lsctables.CoincInspiralTable.get_table(coinc_file)
		sngltable = lsctables.SnglInspiralTable.get_table(coinc_file)

		coinc_time = coinctable[0].end_time + 10.**-9 * coinctable[0].end_time_ns
		for r in sngltable:
			snrs[r.ifo] = r.snr

		# get inj SNR information
		if key == 'noninj':
			inj_time = None
			inj_snrs = defaultdict(lambda: 0)
			source = key
		else:
			inj_time, inj_snrs, dec_snr, source = process_injection(coinc_file, on_ifos)

		return snrs, inj_snrs, dec_snr, source, inj_time

def process_injection(xmldoc, on_ifos):
	inj_snrs = defaultdict(lambda: None)

	simtable = lsctables.SimInspiralTable.get_table(xmldoc)

	time = simtable[0].geocent_end_time + 10.**-9 * simtable[0].geocent_end_time_ns
	inj_snrs['H1'] = simtable[0].alpha4
	inj_snrs['L1'] = simtable[0].alpha5
	inj_snrs['V1'] = simtable[0].alpha6

	source = utils.source_tag(simtable)

	dec_snr = utils.decisive_snr(inj_snrs, on_ifos)

	return time, inj_snrs, dec_snr, source

def main():
	opts, args = parse_command_line()
	
	tag = opts.tag
	topics = opts.input_topic
	datasource = opts.data_source
	far_threshold = float(opts.far_threshold)
	
	# set up producer
	client = kafka.Client(f'kafka://{tag}@{opts.kafka_server}')
	
	# set up logging
	utils.set_up_logger(opts.verbose)
	
	# create a job service using cronut
	app = App('inj_missed_found', broker=f'kafka://{tag}_inj_missed_found@{opts.kafka_server}')

	# subscribes to a topic
	@app.process(topics)
	def process(message): 
		mtopic = message.topic().split('.')[-1]
		mpipeline = message.topic().split('.')[0]
		mkey = utils.parse_msg_key(message)
	
		if mtopic == 'missed_inj':
			# these are injections that never got an event from the search
			# so they are automatically missed
			is_recovered = 'missed'
	
			injection = json.loads(message.value())
			sim_file = utils.load_xml(injection['sim'])
			on_ifos = injection['onIFOs']
	
			time, inj_snrs, dec_snr, source = process_injection(sim_file, on_ifos)
	
			logging.info(f'{mpipeline}: {source} injection from time {time} {is_recovered.upper()}: no associated event message received.')
	
			output = construct_output(injected_snrs = inj_snrs, time = time, decisive_snr = dec_snr)
			for topic in output:
				client.write(f'{mpipeline}.{tag}.testsuite.{topic}', output[topic], tags = [source, is_recovered])
				logging.info(f'Sent msg to: {topic} with tags: {source}, {is_recovered}')
	
		# otherwise, grab the coinc file, far, snr and time from the message
		# unpack data from the message, parse event info, and produce an output message
		elif mtopic == 'events':
			event = json.loads(message.value())
	
			coinc_file = utils.load_xml(event['coinc'])
			far = event['far']
			time = event['time'] + event['time_ns'] * 10**-9.
			ifos = event['onIFOs']
	
			# determine missed or found by getting the far of the recovered event
			is_recovered = 'found' if far < far_threshold else 'missed'
	
			snrs, inj_snrs, dec_snr, source, inj_time = process_event(coinc_file, ifos, mkey)
			logging.info(f'{mpipeline}: {source} event from time {time} {is_recovered.upper()}: far = {far}.')
	
			# send messages to output topics
			output = construct_output(injected_snrs = inj_snrs, recovered_snrs = snrs, time = time, decisive_snr = dec_snr)
			for topic in output:
				client.write(f'{mpipeline}.{tag}.testsuite.{topic}', output[topic], tags = [source, is_recovered])
				logging.info(f'Sent msg to: {topic} with tags: {source}, {is_recovered}')

	
	# start up
	logging.info('Starting up...')
	app.start()

if __name__ == '__main__':
	main()



#!/usr/bin/env python3

from optparse import OptionParser
import os
import sys
import json
import logging
import copy

from collections import defaultdict, deque

from confluent_kafka import Producer
from cronut import App
from cronut.utils import uriparse
from io import BytesIO

from ligo.lw import lsctables
from ligo.lw.utils import load_fileobj

from lal import GPSTimeNow

from ligo.skymap.bayestar import localize
from ligo.skymap.io import events as LIGOSkymapEvents
from ligo.skymap.io import read_sky_map
from ligo.skymap.postprocess import crossmatch

from astropy.coordinates import SkyCoord
from astropy.table import Table

from ligo.scald.io import kafka

from gw.lts import utils
from gw.lts.utils.gracedb_helper import GraceDbHelper

def parse_command_line():
	parser = utils.add_general_opts()
	parser.add_option('--output', help = 'Output directory to write skymap fits files to.')
	parser.add_option('--gdb-skymaps', action='store_true', default = False, help = 'Use skymaps from GraceDB instead of calculating them manually.')
	opts, args = parser.parse_args()

	return opts, args

class Skymap(object):
	def __init__(self, options):
		self.tag = options.tag
		self.kafka_server = options.kafka_server
		self.topics = options.input_topic
		self.output_dir = options.output

		self.gdb_skymaps = options.gdb_skymaps

		if self.gdb_skymaps:
			self.gracedb_helper = GraceDbHelper(options.group)
		
		# set up producer
		self.client = kafka.Client(f'kafka://{self.tag}@{self.kafka_server}')

		self.events = deque(maxlen=10)
	
		# create a job service using cronut
		self.app = App('skymap', broker=f'kafka://{self.tag}_skymap@{self.kafka_server}')
		
		# subscribes to a topic
		@self.app.process(self.topics)
		def process(message): 
			mtopic = message.topic().split('.')[-1]
			mpipeline = message.topic().split('.')[0]
			mkey = utils.parse_msg_key(message)
			logging.info(f'Read message from {mpipeline} {mtopic}.')
		
			# parse event info
			event = json.loads(message.value())
			time = event['time'] + event['time_ns'] * 10**-9.
			event.update({'pipeline': mpipeline})
		
			# process the eventt - get a skymap and calculate
			# searched area and probability, send messages to
			# kafka
			response = self.process_event(event)
			if not response:
				# keep track of events that failed
				# to get a skymap on the first try
				# when getting skymaps from gracedb, this
				# can happen if the skymap isnt uploaded
				# immediately
				times = [e["time"] for e in self.events]
				if not event["time"] in times:
					self.events.append(event)
		
			# iterate over events and try again to grab a
			# skymap for each one. On success, remove the
			# event from the deque
			for e in copy.deepcopy(self.events):
				response = self.process_event(e)
				if response:
					self.events.remove(e)


	def start(self):
		# start up
		logging.info('Starting up...')
		self.app.start()


	def process_event(self, event):
		# either download skymap from gracedb or
		# generate one with bayestar
		if self.gdb_skymaps:
			file = self.gracedb_helper.query_file(event['uid'], 'bayestar.multiorder.fits', outpath = self.output_dir, tag = self.tag)
			if file:
				skymap = Table.read(os.path.join(self.output_dir, f'{self.tag}-{event["uid"]}.fits'))
			else:
				skymap = None
		else:
			skymap = self.make_skymap(event)
	
		if skymap:
			output = {}
			time = event['time']
			pipeline = event['pipeline']

			# get right ascension and declination
			# from sim inspiral table
			coinc = utils.load_xml(event['coinc'])
			simtable = lsctables.SimInspiralTable.get_table(coinc)
			ra, dec = simtable[0].ra_dec
	
			# use SkyCoord and crossmatch to get pvalues
			loc = SkyCoord(ra=ra, dec=dec, unit="rad")
			p = crossmatch(skymap, loc).searched_prob
			deg2 = crossmatch(skymap, loc).searched_area
	
			logging.debug(f'Searched probability: {p} | searched area: {deg2}')
	
			# construct output dict and send messages to kafka
			output['searched_prob'] = {
				'time': [ time ],
				'data': [ p ]
			}
			output['searched_area'] = {
				'time': [ time ],
				'data': [ deg2 ]
			}
	
			for key, value in output.items():
				self.client.write(f'{pipeline}.{self.tag}.testsuite.{key}', value, tags = key)
				logging.info(f'Sent msg to: {pipeline}.{self.tag}.testsuite.{key}')
	
			return True
	
		else:
			return False
	
	def make_skymap(self, event):
		skymap = None
		coinc_obj = BytesIO(event['coinc'].encode())
		# make a copy of coinc file to avoid I/O error in events.ligolw.open()
		psd_obj = copy.copy(coinc_obj)
	
		event_source = LIGOSkymapEvents.ligolw.open(coinc_obj, psd_file=psd_obj, coinc_def=None)
	
		if len(event_source) > 1:
			logging.info('Warning: len(event_source) > 1, defaulting to use the first event.')
	
		# produce the skymap
		for event_id, event in event_source.items():
			skymap = localize(event)
			# break out of the loop after computing just one skymap
			# in case there are multiple rows in coinc inspiral we just
			# go with the first one
			break
	
		return skymap
	

def main():
	opts, args = parse_command_line()

	# set up logging
	utils.set_up_logger(opts.verbose)

	# make a dir for skymaps
	if not os.path.exists(opts.output):
		os.makedirs(opts.output)

	# start up the processor
	processor = Skymap(opts)
	processor.start()

if __name__ == '__main__':
	main()



#!/usr/bin/env python3

from optparse import OptionParser
import os
import sys
import json
import logging

from collections import defaultdict, deque

from confluent_kafka import Producer
from cronut import App
from cronut.utils import uriparse

from ligo.lw import lsctables

from ligo.scald.io import kafka

from gw.lts import utils

def parse_command_line():
	parser = utils.add_general_opts()

	parser.add_option('--ifo', help = 'The interferometer data to get recovered injections from. Required.')
	parser.add_option('--input-params', default = [], action = 'append', help = 'Parameters from sim inspiral table to test recovery. Can be given multiple times.')
	opts, args = parser.parse_args()

	return opts, args

def find_sngl_row(ifo, table):
	for i, row in enumerate(table):
		if table.getColumnByName('ifo')[i] == ifo:
			return i
	return None

def end_time_accuracy(simtable, sngltable, index):
	try:
		injected_time = simtable.getColumnByName('geocent_end_time')[0] + 10.**-9. * simtable.getColumnByName('geocent_end_time_ns')[0]
		recovered_time = sngltable.getColumnByName('end_time')[index] + 10.**-9. * sngltable.getColumnByName('end_time_ns')[index]

		# calculate difference in ms
		diff = (recovered_time - injected_time) * 10**3.
	except Exception as e:
		logging.debug(f'Error getting end time difference: {e}')
		diff = None

	return diff

def fractional_param_accuracy(param, ifo, simtable, sngltable):
	# find this ifo in the sngl table,
	# if this ifo is not  in the sngl table,
	# dont bother continuing
	idx = find_sngl_row(ifo, sngltable)
	if idx is None:
		return None

	# end time is handled differently because
	# we just want the difference in ms
	if param == 'end_time':
		accuracy = end_time_accuracy(simtable, sngltable, idx)
		return accuracy

	inj_param = None
	rec_param = None
	frac_accuracy = None

	# first get the injected parameter value 
	# from the sim inspiral table
	try:
		inj_param = simtable.getColumnByName(param)[0]
	except Exception as e:
		logging.debug(f'Error getting {param} from sim table: {e}')
	if inj_param is None:
		# some parameters can be computed if they are missing
		if param == 'eta':
			mass1 = simtable.getColumnByName('mass1')[0]
			mass2 = simtable.getColumnByName('mass2')[0]
			inj_param = utils.eta_from_m1_m2(mass1, mass2)
		elif param == 'mchirp':
			mass1 = simtable.getColumnByName('mass1')[0]
			mass2 = simtable.getColumnByName('mass2')[0]
			inj_param = utils.mchirp_from_m1_m2(mass1, mass2)

	# get the recovered parameter value
	# from the sngl inspiral table
	try:
		rec_param = sngltable.getColumnByName(param)[idx]
	except Exception as e:
		logging.debug(f'Error getting {param} from sngl table: {e}')

	# calculate fractional accuracy
	if rec_param and inj_param:
		frac_accuracy = 1. - (rec_param - inj_param) / inj_param
		logging.debug(f'{param}: rec: {rec_param} | inj: {inj_param} | accuracy: {frac_accuracy}')

	return frac_accuracy

def main():
	opts, args = parse_command_line()
	
	# sanity check input options
	required_opts = ['ifo', 'tag', 'input_params']
	for r in required_opts:
		if not getattr(opts, r):
			raise ValueError(f'Missing option: {r}.')
	
	ifo = opts.ifo
	tag = opts.tag
	
	# set up producer
	client = kafka.Client(f'kafka://{tag}@{opts.kafka_server}')
	
	# set up logging
	utils.set_up_logger(opts.verbose)
	
	# create a job service using cronut
	app = App('inj_accuracy', broker=f'kafka://{tag}_{ifo}_inj_accuracy@{opts.kafka_server}')
	
	# subscribes to a topic
	@app.process(opts.input_topic)
	def process(message): 
		mtopic = message.topic().split('.')[-1]
		mpipeline = message.topic().split('.')[0]
		mkey = utils.parse_msg_key(message)
		logging.info(f'Read message from input {mtopic}.')
	
		# dont process noninjection events
		if not mkey == 'noninj':
			# parse event info
			event = json.loads(message.value())
	
			coinc_file = utils.load_xml(event['coinc'])
			time = event['time'] + event['time_ns'] * 10**-9.
	
			# load sim inspiral and sngl inspiral tables
			SimInspiralTable = lsctables.SimInspiralTable.get_table(coinc_file)
			SnglInspiralTable = lsctables.SnglInspiralTable.get_table(coinc_file)
	
			for param in opts.input_params:
				# grab info from coinc file
				acc = fractional_param_accuracy(param, ifo, SimInspiralTable, SnglInspiralTable)
	
				# produce output message
				if acc is not None:
					output = {
							'time': [ float(time) ],
							'data': [ acc ]
					}
	
					client.write(f'{mpipeline}.{tag}.testsuite.{ifo}_{param}', output)
					logging.info(f'Sent msg to: {mpipeline}.{tag}.testsuite.{ifo}_{param}')
	# start up
	logging.info('Starting up...')
	app.start()

if __name__ == '__main__':
	main()



#!/usr/bin/env python3

from optparse import OptionParser
import os
import sys
import json
import logging

from collections import defaultdict, deque

from confluent_kafka import Producer
from cronut import App
from cronut.utils import uriparse

from ligo.lw import lsctables

from lal import GPSTimeNow

from ligo.scald.io import kafka

from gw.lts import utils

def parse_command_line():

	parser = utils.add_general_opts()
	opts, args = parser.parse_args()

	return opts, args

def main():
	opts, args = parse_command_line()
	
	tag = opts.tag
	datasource = opts.data_source
	
	# set up Kafka client
	client = kafka.Client(f'kafka://{tag}@{opts.kafka_server}')
	
	# set up logging
	utils.set_up_logger(opts.verbose)
	
	# create a job service using cronut
	app = App('latency', broker=f'kafka://{tag}_latency@{opts.kafka_server}')
	
	# subscribes to a topic
	@app.process(opts.input_topic)
	def process(message): 
		mtopic = message.topic().split('.')[-1]
		mpipeline = message.topic().split('.')[0]
		mkey = utils.parse_msg_key(message)
		logging.debug(f'Read message from {mpipeline} {mtopic}.')
	
		# dont process noninjection events
		if not mkey == 'noninj':
			# parse event info
			event = json.loads(message.value())
	
			time = event['time'] + event['time_ns'] * 10**-9.
			latency = event['latency']
	
			if latency:
				# construct output message
				output = {
						'time': [ float(time) ],
						'data': [ float(latency) ]
				}
	
				# send message to output topics
				client.write(f'{mpipeline}.{tag}.testsuite.latency', output)
				logging.info(f'Sent msg to: {mpipeline}.{tag}.testsuite.latency')

	# start up
	logging.info('Starting up...')
	app.start()

if __name__ == '__main__':
	main()



#!/usr/bin/env python

import logging
import sys
import os
import tempfile
import itertools
from optparse import OptionParser

import yaml
from yaml.loader import SafeLoader

import htcondor
from htcondor import dags

from gw.lts import utils
from gw.lts import dags

def parse_command_line():
	parser = OptionParser()
	parser.add_option('-v', '--verbose', action='store_true', help='Be verbose.')
	parser.add_option('-c', '--config', help = 'Path to config.yml.')
	parser.add_option('-n', '--name', default = 'test_suite', help = 'Name for the DAG file.')
	opts, args = parser.parse_args()

	return opts, args


class DAG(object):
	def __init__(self, config_path, name, verbose):
		self.jobs = {
				'send_inj_stream': dags.send_inj_stream_layer,
				'inspinjmsg_find': dags.inspinjmsg_find_layer,
				'igwn_alert_listener': dags.igwn_alert_listener_layer,
				'inj_missed_found': dags.inj_missed_found_layer,
				'vt': dags.vt_layer,
				'latency': dags.latency_layer,
				'p_astro': dags.p_astro_layer,
				'skymap': dags.skymap_layer,
				'snr_consistency': dags.snr_consistency_layer,
				'inj_accuracy': dags.inj_accuracy_layer,
				'scald_metric_collector': dags.collect_metrics_layer,
				'scald_event_collector' : dags.scald_inj_table_events_layer
		}

		self.config_path = config_path
		self.config = self.load_config(self.config_path)
		self.retries = '3'

		self.dag_name = name
		self.dag = htcondor.dags.DAG()

		self.verbose = verbose

	def load_config(self, config_path):
		config = {}
	
		with open(config_path, 'r') as f:
			config = yaml.load(f, Loader=SafeLoader)
	
		return config

	def create_dag(self):
		if self.verbose: print('Building DAG...')
		opts = self.condor_opts()

		# add jobs
		for job in self.config['jobs']:
			self.dag = self.jobs[job](self.dag, self.config, opts)

		# add a scald metric collector job for each datasource
		self.dag = self.jobs['scald_metric_collector'](self.dag, self.config, self.config_path, opts)

		# add a scald event collector job
		self.dag = self.jobs['scald_event_collector'](self.dag, self.config, self.config_path, opts)

		if self.verbose: print(f'Writing DAG out to file: {self.dag_name}.dag')
		htcondor.dags.write_dag(self.dag, os.getcwd(), dag_file_name=f'{self.dag_name}.dag')

		return self.dag

	def write_script(self):
		if self.verbose: print(f'Writing script out to file: {self.dag_name}.sh')
		with open(f'{self.dag_name}.sh', 'w') as f:
			jobnames = []
			for layer in self.dag.walk():
				executable = layer.submit_description['executable']
				args = layer.submit_description['arguments']
				jobname = executable.split('/')[-1]
	
				jobnames.extend([jobname])
				if jobname in jobnames:
					jobname = jobname + f'_{int(jobnames.count(jobname)):04d}'
	
				print(f'# Job {jobname}', file=f)
				print(executable + ' ' + args + '\n', file=f)

	def condor_opts(self):
		# set up condor logging
		tempfile.tempdir = os.environ['TMPDIR']
		fd, logfile = tempfile.mkstemp()
		with os.fdopen(fd, 'w') as f:
			pass
	
		# set up general condor options across all jobs
		opts = {
			'want_graceful_removal': True,
			'getenv': True,
			'kill_sig': 15,
			'log': logfile,
			'notification': 'never',
			}
		if self.config["singularity-image"]:
			opts.update({
				'MY.SingularityImage': f'\"{self.config["singularity-image"]}\"'
			})
		opts.update(self.config['condor'])

		return opts

def main():
	# parse command line options
	opts, args = parse_command_line()

	if not opts.config: 
		raise Exception('You need to specify a config file.')

	# build the dag
	dag_generator = DAG(opts.config, opts.name, opts.verbose)

	dag = dag_generator.create_dag()
	dag_generator.write_script()
	
	# make a directory for logs
	if not os.path.exists('logs'): os.mkdir('logs')

	print('Done.')

if __name__ == '__main__':
	main()



import logging
from io import StringIO, BytesIO
from ligo.lw import ligolw
from ligo.lw import lsctables
from ligo.lw import utils as ligolw_utils

from ligo.segments import segment
from ligo.scald.utils import floor_div

import yaml
from yaml.loader import SafeLoader

from optparse import OptionParser

class LIGOLWContentHandler(ligolw.LIGOLWContentHandler):
    pass

lsctables.use_in(LIGOLWContentHandler)

all_sngl_rows = ("process:process_id", "ifo", "search", "channel", "end_time", "end_time_ns", "end_time_gmst", "impulse_time", "impulse_time_ns", "template_duration", "event_duration", "amplitude", "eff_distance", "coa_phase", "mass1", "mass2", "mchirp", "mtotal", "eta", "kappa", "chi", "tau0", "tau2", "tau3", "tau4", "tau5", "ttotal", "psi0", "psi3", "alpha", "alpha1", "alpha2", "alpha3", "alpha4", "alpha5", "alpha6", "beta", "f_final", "snr", "chisq", "chisq_dof", "bank_chisq", "bank_chisq_dof", "cont_chisq", "cont_chisq_dof", "sigmasq", "rsqveto_duration", "Gamma0", "Gamma1", "Gamma2", "Gamma3", "Gamma4", "Gamma5", "Gamma6", "Gamma7", "Gamma8", "Gamma9", "spin1x", "spin1y", "spin1z", "spin2x", "spin2y", "spin2z", "event_id")

all_coinc_rows = ("coinc_event:coinc_event_id", "combined_far", "end_time", "end_time_ns", "false_alarm_rate", "ifos", "mass", "mchirp", "minimum_duration", "snr")

all_coinc_event_rows = ("coinc_definer:coinc_def_id", "coinc_event_id", "instruments", "likelihood", "nevents", "process:process_id", "time_slide:time_slide_id")

all_coinc_map_rows = {"coinc_event:coinc_event_id", "event_id", "table_name"}

all_process_rows = ("comment", "cvs_entry_time", "cvs_repository", "domain", "end_time", "ifos", "is_online", "jobid", "node" , "process_id", "program", "start_time", "unix_procid", "username", "version")


def add_general_opts():
	parser = OptionParser()
	parser.add_option('--data-source', metavar = 'string', action='append', help = 'Source of test suite data. Options: fake-data, gstlal, mbta, pycbc.')
	parser.add_option('--tag', help = 'The tag used to uniquely identify the analysis you wish to process metrics from. Used as Kafka group ID.')
	parser.add_option('--kafka-server', metavar = 'string', help = 'Sets the url for the kafka broker.')
	parser.add_option('--analysis-dir', metavar='path', help = '')
	parser.add_option('--inj-file', metavar='file', help='')
	parser.add_option('--input-topic', metavar = 'string', action='append', help = 'The Kafka topic(s) to subscribe to.')
	parser.add_option('--group', metavar='string', default = 'gracedb-playground', help = 'GraceDb group to use. Valid options are gracedb, gracedb-playground, and gracedb-test.')
	parser.add_option('--verbose', default=False, action="store_true", help = 'Be verbose.')

	return parser

def set_up_logger(verbose):
	logging.getLogger()
	if verbose:
		log_level = logging.DEBUG
	else:
		log_level = logging.INFO

	handler = logging.StreamHandler()
	formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
	handler.setFormatter(formatter)
	logging.getLogger('').addHandler(handler)
	logging.getLogger('').setLevel(log_level)

def load_xml(f, gz=False):
	if isinstance(f, str):
		f = BytesIO(f.encode("utf-8"))
	xmldoc = ligolw_utils.load_fileobj(f, contenthandler = LIGOLWContentHandler, gz=gz)

	return xmldoc

def load_filename(f, gz=True):
	xmldoc = ligolw_utils.load_filename(f, contenthandler = LIGOLWContentHandler, gz=gz)

	return xmldoc

def event_window(t):
	"""
	returns the event window representing the event
	"""
	dt = 0.2
	return segment(floor_div(t - dt, 0.5), floor_div(t + dt, 0.5) + 0.5)

def get_ifos(sngltable):
	ifos = ''
	for row in sngltable:
		if not row.ifo in ifos:
			ifos += str(row.ifo)

	return ifos

def source_tag(simtable):
	mass1 = simtable[0].mass1
	mass2 = simtable[0].mass2
	cutoff = 3 # minimum BH mass

	if mass1 < cutoff and mass2 < cutoff: source = 'BNS'
	elif mass1 >= cutoff and mass2 >= cutoff: source = 'BBH'
	else: source = 'NSBH'
	return source

def parse_msg_key(message):
	try:
		key = message.key().decode("utf-8")
	except AttributeError:
		key = None
	return key

def find_nearest_msg(msgs, t):
	# get injections close to the event end time
	delta = 1.0

	# construct a list of tuples (msg, Delta_t) where Delta_t is the difference in time
	# between this msg and the input time, sort by Delta_t, and take the first element
	# in the list which corresponds to the msg closest to the input time
	nearest_msg = None
	try:
		near_msgs = list((msg, abs(msg['time'] - t)) for msg in msgs if t - delta <= msg['time'] <= t + delta)
		if near_msgs:
			nearest_msg = sorted(near_msgs, key = lambda x: x[1])[0][0]

			logging.debug(f'Time to search for: {t} | Nearest msg time: {nearest_msg["time"]}')
	except Exception as e:
		logging.debug(f'Error: {e}')
	return nearest_msg

def decisive_snr(sngl_snrs, ifos):
	ifos = ifos.split(',')
	sngl_snrs = [sngl_snrs[ifo] for ifo in ifos]

	# if no ifos, assume all were off
	if len(ifos) == 0:
		return 0
	# for single time, use the most sensitive IFO
	elif len(ifos) == 1:
		return sorted(sngl_snrs, reverse=True)[0]
	# for coinc time, use the SNR of the second most sensitive IFO
	elif len(ifos) >= 2:
		return sorted(sngl_snrs, reverse=True)[1]

def eta_from_m1_m2(m1, m2):
	# symmetric mass ratio from component masses
	m1 = float(m1)
	m2 = float(m2)
	return (m1 * m2) / (m1 + m2)**2.

def mchirp_from_m1_m2(m1, m2):
	# chirp mass from component masses
	m1 = float(m1)
	m2 = float(m2)
	return (m1 * m2)**(3./5.) / (m1 + m2)**(1./5.)

def preferred_event_func(func):
	if func == 'max' or func == 'latest':
		return max
	elif func == 'min' or func == 'first':
		return min
	else:
		raise NotImplementedError

def parse_dq_topic(topic):
	pipeline, tag, topic = topic.split('.')
	ifo, topic = topic.split('_')

	return pipeline, tag, ifo, topic




import os
import setuptools
from pathlib import Path

long_description = (Path(__file__).parent / "README.md").read_text()

install_requires = [
	'confluent-kafka>=1.7.0',
	'pyyaml',
	'htcondor',
	'htcondor-dags',
	'python-ligo-lw==1.7.1',
	'lalsuite==6.82',
	'cronut>=0.1.0',
	'ligo-scald>=0.8.0',
	'ligo.skymap',
	'igwn-alert',
	'hop-client',
]

setuptools.setup(
	name = 'gw-lts',
	description = 'Gravitational Wave Low-Latency Test Suite',
	long_description = long_description,
	long_description_content_type='text/markdown',
	version = '0.0.4',
	author = 'Becca Ewing',
	author_email = 'rebecca.ewing@ligo.org',
	url = 'https://git.ligo.org/rebecca.ewing/gw-lts.git',
	license = 'MIT',

	packages = ['gw', 'gw.lts', 'gw.lts.utils', 'gw.lts.dags'],

	entry_points = {
		'console_scripts': [
			'igwn-alert-listener=gw.lts.igwn_alert_listener:main',
			'send-inj-stream=gw.lts.send_inj_stream:main',
			'inspinjmsg-find=gw.lts.inspinjmsg_find:main',
			'inj-missed-found=gw.lts.inj_missed_found:main',
			'vt=gw.lts.vt:main',
			'snr-consistency=gw.lts.snr_consistency:main',
			'inj-accuracy=gw.lts.inj_accuracy:main',
			'p-astro=gw.lts.p_astro:main',
			'skymap=gw.lts.skymap:main',
			'latency=gw.lts.latency:main',
			'likelihood=gw.lts.likelihood:main',
			'test-suite-workflow=gw.lts.dags.test_suite_workflow:main',
		],
	},

	data_files = [('etc', ['etc/example_config.yml', 'etc/example_fake_data_config.yml']),
		('injections', ['injections/fake_data_injections.xml.gz'])
	],

	python_requires = '>=3.6.*',
	install_requires = install_requires,
	zip_safe = False,

	classifiers = [
		'Development Status :: 3 - Alpha',
		'Programming Language :: Python',
		'Programming Language :: Python :: 3',
		'Intended Audience :: Science/Research',
		'Natural Language :: English',
	],

)



#!/usr/bin/env python3

from optparse import OptionParser
import logging
import json
import os
import io
import sys
import copy
import numpy
import random
from time import sleep
from collections import defaultdict, OrderedDict

from ligo.lw import ligolw
from ligo.lw import lsctables
from ligo.lw import utils as ligolw_utils
from ligo.lw.param import Param
from ligo.lw.utils.process import register_to_xmldoc
from ligo.skymap.tool import bayestar_realize_coincs
from ligo.skymap.bayestar import filter as bayestar_filter

from confluent_kafka import Producer, Consumer

import lal
from lal import GPSTimeNow, LIGOTimeGPS, GreenwichMeanSiderealTime
from lal import MSUN_SI, PC_SI
import lal.series
import lalsimulation

from ligo.scald.io import kafka

from gw.lts import utils
from gw.lts.utils import cosmology_utils as cutils
from gw.lts.utils import pastro_utils

class LIGOLWContentHandler(ligolw.LIGOLWContentHandler):
	pass

lsctables.use_in(LIGOLWContentHandler)

# define input opts
def parse_command_line():
	parser = utils.add_general_opts()
	parser.add_option('--time-offset', type = int, help = 'Time offset to shift injections by. Not used if data-source is fake-data, required otherwise.. If using mdc injections this should be non-zero number that corresponds to the time h(t) is shifted by. For gstlal injections from disk, this should be 0.')
	parser.add_option('--psd-file', metavar = 'file', help = 'Path to reference PSD file.')
	parser.add_option('--track-psd', action='store_true', default = False, help = 'Dynamically update PSD. If given, kafka-server and input-topic are required. Default is False.')
	parser.add_option('--track-segments', action='store_true', default = False, help = 'Track IFO states. If given, kafka-server and input-topic are required. Default is False.')
	parser.add_option('--ifo', action = 'append', help = 'IFOs to use. Can be given more than once.')
	parser.add_option('--far-threshold', type = float, default = 2.315e-5, help = 'Set the FAR threshold for sending coincs. Used if data-source is fake-data. Default is 2 per day.')
	parser.add_option('--inj-rate', type = float, default = 20., help = 'Rate to send injection messages in the fake-data scheme. Default is 20 seconds.')
	parser.add_option('--output-coinc', metavar = 'dir', default = 'output_files', help='the directory to output coinc files to')
	opts, args = parser.parse_args()

	return opts, args

class SendInjStream(object):
	def __init__(self, options):
		logging.info('Setting up injection stream...')

		self.verbose = options.verbose
		self.tag = options.tag
		self.datasources = options.data_source
		self.kafka_server = options.kafka_server
		self.time_offset = options.time_offset
		self.ifo_name = options.ifo
		self.inj_rate = options.inj_rate
		self.far_threshold = options.far_threshold
		self.output_coinc = options.output_coinc
		self.track_segments = options.track_segments
		self.track_psd = options.track_psd
		self.send_fake_coincs = True if 'fake-data' in self.datasources else False

		### set up kafka consumer and producer
		self.producer = kafka.Client(f'kafka://{self.tag}@{self.kafka_server}')
		
		if options.input_topic:
			kafka_settings = {
			'bootstrap.servers': self.kafka_server,
			'group.id': self.tag,
			'message.max.bytes': 5242880 # 5 MB
		}
			self.client = Consumer(kafka_settings)
			self.client.subscribe([topic for topic in options.input_topic])

		### set up psds and segments
		self.new_psds = {}
		self.psds = self.load_psd(options.psd_file)
		self.psds_dict = {key: bayestar_filter.InterpolatedPSD(bayestar_filter.abscissa(psd), psd.data.data)
			for key, psd in self.psds.items() if psd is not None}

		if options.track_segments:
			self.segments = defaultdict(lambda: OrderedDict())

		### additional set up for the fake data configuration
		if self.send_fake_coincs:
			self.fake_data_setup()

		# load sorted sim inspiral table
		self.simtable = self.load_sim_table(options.inj_file)

	def start(self):
		logging.info('Starting injection stream...')
		# step through sim table row by row, update sim col values as necessary,
		# send message to inj_stream, then if data-source is fake-data, 
		# generate a coinc file and send a message to events topic
		while self.simtable:
			# remove and return the oldest/first inj in the table
			thisrow = self.simtable.pop(0)
		
			if self.send_fake_coincs:
				# sleep for inj cadence
				sleep(self.inj_rate)
		
				timenow = float(GPSTimeNow())
				time_offset = timenow - thisrow.geocent_end_time
				injection_time = thisrow.geocent_end_time + time_offset
		
			else:
				timenow = float(GPSTimeNow())
				time_offset = self.time_offset
				injection_time = thisrow.geocent_end_time + time_offset
		
				# if the injection is already passed, skip it
				if timenow - (injection_time) >= 5 * 60.:
					logging.debug('Skipping old injection')
					continue
		
				# sleep until its close to the next inj time
				time_to_sleep = (injection_time) - timenow
				if time_to_sleep > 0.:
					logging.debug('Sleeping until next injection...')
					sleep(time_to_sleep)
		
			logging.debug(f'Processing injection at coalescence time: {injection_time}')

			# track segments and psds
			if self.track_segments:
				self.segments_tracker(injection_time)
			elif self.track_psd:
				new_psd_msgs, _ = self.pull_dq_messages()
				self.new_psds.update(new_psd_msgs)
		
			# once we have the state vector segments that we need,
			# update the psds just once
			if self.track_psd and self.new_psds:
				self.track_psds()

			# proceed to generate the correct sim row and send
			# an output message kafka
			outxml, output_simtable = self.new_sim_xml()
			output_row = copy.copy(thisrow)

			# if time offset is nonzero, shift the times, longitude
			# and re-calculate the inj snrs
			if time_offset:
				output_row = self.shift_times(output_row, time_offset)
				inj_snrs = self.calc_inj_snrs(output_row)

				# add inj snrs to appropriate cols in output_row 
				output_row.alpha4 = inj_snrs['H1']
				output_row.alpha5 = inj_snrs['L1']
				output_row.alpha6 = inj_snrs['V1']

				logging.debug(f"SNRs: H1: {output_row.alpha4} | L1: {output_row.alpha5} | V1: {output_row.alpha6}")

			# construct sim table
			output_simtable.append(output_row)
			outxml.childNodes[-1].appendChild(output_simtable)

			sim_msg = io.BytesIO()
			ligolw_utils.write_fileobj(outxml, sim_msg, gz = False)

			# get state vector at time of this injection
			# in fake data configuration, lets just assume all IFOs
			# are always on
			onIFOs = []
			if self.send_fake_coincs:
				onIFOs = ['H1', 'L1', 'V1']
			# FIXME: we are assuming that no state info for a given time
			# means that the IFO was off. This is a bad idea and we should
			# really fix it to make sure that we receive explicit 0 states 
			# at times when the IFOs actually were off.
			if self.track_segments:
				for ifo, states in self.segments.items():
					try:
						ifo_on = states[injection_time]
						if ifo_on:
							onIFOs.append(ifo)
					except KeyError:
						near_times = []
						for time, state in states.items():
							if injection_time - 0.5 < time < injection_time + 0.5:
								near_times.append((time, state))

						ifo_on = near_times and all(s[1] for s in near_times)
						if ifo_on:
							onIFOs.append(ifo)
			logging.debug(f'on IFOs: {onIFOs}')

			# construct output json packet
			output = {
					'sim': sim_msg.getvalue().decode(),
					'onIFOs': (',').join(onIFOs)
			}

			# output msgs to kafka
			for datasource in self.datasources:
				self.producer.write(f'{datasource}.{self.tag}.inj_stream', output)
				logging.info(f'Sent msg to: {datasource}.{self.tag}.inj_stream')

			if not self.send_fake_coincs:
				outxml.unlink()
				continue

			# otherwise, proceed to send the coinc message
			trigger = copy.copy(output_row)
			outxml.unlink()

			coincfar = self.snr_to_far_map([trigger.alpha4, trigger.alpha5, trigger.alpha6])

			# propduce a fake coinc if the far passes the threshold
			if coincfar < self.far_threshold:
				logging.debug('Sending a coinc trigger...')
				resp = self.produce_coinc_output(trigger, coincfar)
			else:
				logging.debug(f'Coinc FAR {coincfar:e} above threshold to send message')

		logging.info('Sent all injections. Exiting ...')

	def load_psd(self, file):
		xmldoc = ligolw_utils.load_filename(file, contenthandler = lal.series.PSDContentHandler)
		return lal.series.read_psd_xmldoc(xmldoc, root_name=None)

	def fake_data_setup(self):
		self.detectors = [lalsimulation.DetectorPrefixToLALDetector(ifo) for ifo in self.ifo_name]
		self.responses = [det.response for det in self.detectors]
		self.locations = [det.location for det in self.detectors]
		
		# the interpolated object below will be used for SNR time series simulation
		self.psds_interp = [self.psds_dict[ifo] for ifo in self.ifo_name]

		### compute distributions required for pastro calculation
		self.p_x_c = pastro_utils.p_x_c(bns=(1.22, .06), nsbh=(6.27, 1.84), bbh=(42.98, 8.11))
		self.p_c = pastro_utils.p_c(self.p_x_c, N_events={"Terrestrial": 0, "BNS": 60800, "NSBH": 48400, "BBH": 60800})

	def load_sim_table(self, file):
		xmldoc = ligolw_utils.load_filename(file, contenthandler = LIGOLWContentHandler)
		simtable = lsctables.SimInspiralTable.get_table(xmldoc)
		simtable.sort(key = lambda row: row.geocent_end_time + 10.**-9. * row.geocent_end_time_ns)

		return simtable

	def segments_tracker(self, injection_time, max_retries = 10):
		have_states = {ifo: False for ifo in self.ifo_name}

		tries = 0
		while not all(have_states.values()) and tries < max_retries:
			for ifo in self.ifo_name:
				# either we have a state corresponding to this injection time
				# or we have later states (for now assuming no state = IFO off)
				have_states[ifo] = self.segments[ifo] and (injection_time in self.segments[ifo].keys() or next(reversed(self.segments[ifo])) > injection_time)

				if have_states[ifo]:
					# move on to check the next ifo
					continue
				else:
					# sleep for the difference in time between the latest state
					# and this injection so that the ll_dq job can catch up
					if self.segments[ifo] and next(reversed(self.segments[ifo])) - injection_time > 0:
						sleep(next(reversed(self.segments[ifo])) - injection_time)

					# pull and store new messages and try again
					new_psd_msgs, new_segs = self.pull_dq_messages()
					self.new_psds.update(new_psd_msgs)
					self.store_segments(new_segs)

					tries += 1
					break

	def store_segments(self, new_segs):
		for ifo in new_segs.keys():
			for time, states in sorted(new_segs[ifo].items()):
				# for each time, we can receive states from each job.
				# take the max over the recored state from each job,
				# ie if at least one job reports that the data was on 
				# at this time, assume it was really on.
				self.segments[ifo].update({time: int(max(states))})
			while len(self.segments[ifo].keys()) >= 500.:
				self.segments[ifo].popitem(last=False)

	def pull_dq_messages(self, num_messages=10000, timeout=0.3):
		psds = {}
		statevectorsegments = defaultdict(lambda: {})
		msgs = self.client.consume(num_messages=num_messages, timeout=0.2)
		for msg in sorted(msgs, key = self.sortfunc, reverse = True):
			if msg and not msg.error():
				pipeline, tag, ifo, topic = utils.parse_dq_topic(msg.topic())

				if topic.endswith("psd"):
					psd = json.loads(msg.value())
					psds.setdefault(ifo, psd)
	
				elif topic.endswith("statevectorsegments"):
					value = json.loads(msg.value())
					time = value['time']
					state = value['data']
					for t, s in zip(time, state):
						# if we're getting segs from each job, there will be duplicates.
						# append each state to a list for a given time
						statevectorsegments[ifo].setdefault(t, []).append(s)
	
		return psds, statevectorsegments

	def track_psds(self):
		for ifo, data in self.new_psds.items():
			# parse psd data
			x = numpy.array(data['freq'])
			y = abs(numpy.array(data['asd']))**2.
	
			# make sure they are the same length
			if len(y) != len(x):
				x = x[:len(y)]
	
			# remove nans
			psd_data = numpy.array([])
			frequency = numpy.array([])
			for f, p in zip(x, y):
				if not numpy.isnan(p):
					psd_data = numpy.append(psd_data, p)
					frequency = numpy.append(frequency, f)
	
			# update psds dict with interpolated psds
			new_psd = lal.CreateREAL8FrequencySeries('new_psd', None, min(frequency), data['deltaF'], 's strain^2', len(psd_data))
			new_psd.data.data = psd_data
			self.psds.update({ifo: new_psd})
	
			self.psds_dict.update({ifo: bayestar_filter.InterpolatedPSD(frequency, psd_data)})
			logging.debug(f'Updated {ifo} PSD.')
	
		self.psds_interp = [self.psds_dict[ifo] for ifo in self.psds.keys()]

	def sortfunc(self, m):
		return m.timestamp()

	def new_sim_xml(self):
		# open a new xml doc
		xml = ligolw.Document()
		xml.appendChild(ligolw.LIGO_LW())
	
		# write a sim inspiral table with a single row corresponding to this injection
		simtable = lsctables.New(lsctables.SimInspiralTable)
		return xml, simtable

	def shift_times(self, row, time_offset):
		# fix gpstimes and RA
		row.geocent_end_time = int(row.geocent_end_time + time_offset)
		row.h_end_time = row.h_end_time + time_offset
		row.l_end_time = row.l_end_time + time_offset
		row.v_end_time = row.v_end_time + time_offset
	
		gmst0 = GreenwichMeanSiderealTime(LIGOTimeGPS(row.geocent_end_time + row.geocent_end_time_ns * 10.**-9.))
		gmst = GreenwichMeanSiderealTime(LIGOTimeGPS(row.geocent_end_time + row.geocent_end_time_ns * 10.**-9. + time_offset))
		dgmst = gmst - gmst0
		row.longitude = row.longitude + dgmst
	
		return row

	def calc_inj_snrs(self, inj):
		snr = dict.fromkeys(self.ifo_name, 0.0)
	
		injtime = inj.geocent_end_time
		f_min = inj.f_lower
		approximant = lalsimulation.GetApproximantFromString(str(inj.waveform))
		sample_rate = 16384.0
		f_max = 1600.
	
		h_plus, h_cross = lalsimulation.SimInspiralTD(
			m1 = inj.mass1*lal.MSUN_SI,
			m2 = inj.mass2*lal.MSUN_SI,
			S1x = inj.spin1x,
			S1y = inj.spin1y,
			S1z = inj.spin1z,
			S2x = inj.spin2x,
			S2y = inj.spin2y,
			S2z = inj.spin2z,
			distance = inj.distance*1e6*lal.PC_SI,
			inclination = inj.inclination,
			phiRef = inj.coa_phase,
			longAscNodes = 0.0,
			eccentricity = 0.0,
			meanPerAno = 0.0,
			deltaT = 1.0 / sample_rate,
			f_min = f_min,
			f_ref = 0.0,
			LALparams = None,
			approximant = approximant
		)
	
		h_plus.epoch += injtime
		h_cross.epoch += injtime
	
		# Compute strain in each detector. If one detector wasn't on, snr will be set to zero.
		for instrument in snr:
			if instrument not in self.psds.keys():
				continue
			h = lalsimulation.SimDetectorStrainREAL8TimeSeries(h_plus, h_cross, inj.longitude, inj.latitude, inj.polarization, lalsimulation.DetectorPrefixToLALDetector(instrument))
			snr[instrument] = lalsimulation.MeasureSNR(h, self.psds[instrument], f_min, f_max)
	
		return snr

	def snr_to_far_map(self, snrs):
		# this is set so that a network SNR 7 event (H1 SNR = L1 SNR = 4.) is recovered with FAR < 2/day
		return 6 * 10**-4. * numpy.exp(- (numpy.sqrt(numpy.linalg.norm([snr for snr in snrs if snr > 4.])))**2. / 2.)

	def produce_coinc_output(self, trigger, coincfar, key=None):
		# build coinc xml doc, calculate p_astro, and produce message
		newxmldoc = self.build_coinc_xml(trigger, coincfar)
		if not newxmldoc:
			return False
	
		coinctable = lsctables.CoincInspiralTable.get_table(newxmldoc)
		coincsnr = coinctable[0].snr
	
		p_astro = self.get_pastro(newxmldoc, coincsnr)
	
		output = self.construct_event_ouput(newxmldoc, p_astro, filekey=key)
	
		# send coinc message to events topic
		logging.info(f"network SNR: {output['snr']} | FAR: {output['far']}")
	
		self.producer.write(f'fake-data.{self.tag}.inj_events', output)
		logging.info(f'Sent msg to: fake-data.{self.tag}.inj_events')
		newxmldoc.unlink()
	
		return True

	def build_coinc_xml(self, row, coincfar):
	
		# instantiate relevant lsctables objects
		newxmldoc = ligolw.Document()
		ligolw_elem = newxmldoc.appendChild(ligolw.LIGO_LW())
		new_process_table = ligolw_elem.appendChild(lsctables.New(lsctables.ProcessTable, columns = utils.all_process_rows))
		new_sngl_inspiral_table = ligolw_elem.appendChild(lsctables.New(lsctables.SnglInspiralTable, columns = utils.all_sngl_rows))
		new_coinc_inspiral_table = ligolw_elem.appendChild(lsctables.New(lsctables.CoincInspiralTable, columns = utils.all_coinc_rows))
		new_coinc_event_table = ligolw_elem.appendChild(lsctables.New(lsctables.CoincTable))
		new_coinc_map_table = ligolw_elem.appendChild(lsctables.New(lsctables.CoincMapTable))
	
		# simulate SNR time series using interpolated psd object
		# measurement_error is set as gaussian but one can switch to no noise by measurement_error="zero-noise"
		bayestar_sim_list = bayestar_realize_coincs.simulate(seed=None, sim_inspiral=row, psds=self.psds_interp, responses=self.responses, locations=self.locations, measurement_error="gaussian-noise", f_low=20, f_high=2048)
	
		# get mass parameters
		mass1 = max(numpy.random.normal(loc=row.mass1, scale=1.), 1.1)
		mass2 = max(numpy.random.normal(loc=row.mass2, scale=1.), mass1)
		mchirp, eta = self.mc_eta_from_m1_m2(mass1, mass2)
	
		snrs = defaultdict(lambda: 0)
		coincsnr = None
	
		# populate process table
		process_row_dict = {k:0 for k in utils.all_process_rows}
		process_row_dict.update({"process_id": 0,
					"program": "gstlal_inspiral", #FIXME
					"comment": ""
					})
		new_process_table.extend([lsctables.ProcessTable.RowType(**process_row_dict)])
	
		# populate sngl table, coinc map table, and SNR timeseriess 
		for event_id, (ifo, (horizon, abs_snr, arg_snr, toa, snr_series)) in enumerate(zip(self.ifo_name, bayestar_sim_list)):
			sngl_row_dict = {k:0 for k in utils.all_sngl_rows}
	
			sngl_row_dict.update({
					"process_id": 0,
					"event_id": event_id,
					"end": toa,
					"mchirp": mchirp,
					"mass1": mass1,
					"mass2": mass2,
					"eta": eta,
					"ifo": ifo,
					"snr": abs_snr,
					"coa_phase": arg_snr
					})
	
			# add to the sngl inspiral table
			new_sngl_inspiral_table.extend([lsctables.SnglInspiralTable.RowType(**sngl_row_dict)])
			snrs[ifo] = abs_snr
	
			coinc_map_row_dict = {"coinc_event_id": 0,
								"event_id": event_id,
								"table_name": "sngl_inspiral",
								}
	
			# add to the coinc map table
			new_coinc_map_table.extend([lsctables.CoincMapTable.RowType(**coinc_map_row_dict)])
	
			# add SNR time series as array objects
			elem = lal.series.build_COMPLEX8TimeSeries(snr_series)
			elem.appendChild(
				Param.from_pyvalue('event_id', event_id))
			ligolw_elem.appendChild(elem)
	
		# calculate coinc SNR, only proceed if above 4
		coincsnr = numpy.linalg.norm([snr for snr in snrs.values() if snr > 4])
		if not coincsnr:
			logging.debug(f'Coinc SNR {coincsnr} too low to send a message.')
			return None
	
		# populate coinc inspiral table
		coinc_row_dict = {col:0 for col in utils.all_coinc_rows}
		coincendtime = row.geocent_end_time
		coincendtimens = row.geocent_end_time_ns
		coinc_row_dict.update({
					"coinc_event_id": 0,
					"snr": coincsnr,
					"mass": row.mass1 + row.mass2,
					"mchirp": row.mchirp,
					"end_time": coincendtime,
					"end_time_ns": coincendtimens,
					"combined_far": coincfar,
					})
		new_coinc_inspiral_table.extend([lsctables.CoincInspiralTable.RowType(**coinc_row_dict)])
	
		# populate coinc event table
		coinc_event_row_dict = {col:0 for col in utils.all_coinc_event_rows}
		coinc_event_row_dict.update({
					"coinc_def_id": 0,
					"process_id": 0,
					"time_slide_id": 0,
					"instruments": "H1,L1,V1",
					"numevents": len(new_sngl_inspiral_table)
					})
		new_coinc_event_table.extend([lsctables.CoincTable.RowType(**coinc_event_row_dict)])
	
		# add psd frequeny series
		lal.series.make_psd_xmldoc(self.psds, ligolw_elem)
	
		return newxmldoc

	def mc_eta_from_m1_m2(self, m1, m2):
		mc = (m1 * m2)**(3./5.) / (m1 + m2)**(1./5.)
		eta = (m1 * m2) / (m1 + m2)**2.
	
		return mc, eta

	def get_pastro(self, xmldoc, rankstat):
		coinctable = lsctables.CoincInspiralTable.get_table(xmldoc)
	
		mchirp = coinctable[0].mchirp
	
		return pastro_utils.p_astro(mchirp, rankstat, self.p_x_c, self.p_c)
	
	def construct_event_ouput(self, xmldoc, p_astro, filekey=None):
		coinctable = lsctables.CoincInspiralTable.get_table(xmldoc)
		time = coinctable[0].end_time
	
		# write coinc file to disk
		file = f'fake_coinc-{int(time)}.xml' if not filekey else f'{filekey}-fake_coinc-{int(time)}.xml'
		ligolw_utils.write_filename(xmldoc, os.path.join(self.output_coinc, file), verbose = self.verbose)
	
		coinc_msg = io.BytesIO()
		ligolw_utils.write_fileobj(xmldoc, coinc_msg, gz = False)
	
		# create json packet
		output = {
			'time': time,
			'time_ns': coinctable[0].end_time_ns,
			'snr': coinctable[0].snr,
			'far': coinctable[0].combined_far,
			'p_astro': json.dumps(p_astro),
			'coinc': coinc_msg.getvalue().decode(),
		}
		return output

def main():
	### parse command line
	opts, args = parse_command_line()

	if opts.track_psd or opts.track_segments:
		if not getattr(opts, 'kafka_server'):
			raise ValueError(f'Must specify --kafka-server when --track-psd or --track-segments is set.')
		if not getattr(opts, 'input_topic'):
			raise ValueError(f'Must specify at least one --input-topic when --track-psd or --track-segments is set.')
	
	try:
		os.mkdir(opts.output_coinc)
	except OSError as error:
		pass
	
	### set up logger
	logger = utils.set_up_logger(opts.verbose)
	
	### initialize and set up
	send_inj_stream = SendInjStream(opts)

	### start the stream
	send_inj_stream.start()

if __name__ == '__main__':
	main()



#!/usr/bin/env python3

import numpy

mc_centers = numpy.logspace(0, 2, 100)
L_centers = numpy.logspace(0, 2.5, 102)

# options: mean and sigma chirp-mass for each signal type 
#          ranking stat threshold

def noise_dist(mc_centers = mc_centers, L_centers = L_centers):
	# make up a noise distribution which is independent of mchirp, p_noise = exp(-L)
	p_x_c0 = numpy.zeros((len(mc_centers), len(L_centers)))
	for i in range(len(mc_centers)): p_x_c0[i,:] = numpy.exp(-L_centers)
	p_x_c0 /= p_x_c0.sum()
	return p_x_c0

# signal distribution is a gaussian in chirp mass defined by some mean and sigma values; 
# ranking stat is used for scaling 

def signal_dist(mean, sigma, mc_centers=mc_centers, L_centers=L_centers):
	p_x_ci = numpy.zeros((len(mc_centers), len(L_centers)))
	for i in range(len(mc_centers)):
		p_x_ci[i,:] = numpy.exp(-(mc_centers[i] - mean)**2./(2.*sigma**2.)) * L_centers**-4.
	p_x_ci /= p_x_ci.sum()
	return p_x_ci


def p_x_c(bns=(1.2, .2), nsbh=(3.0, .4), bbh=(50., 10.)):
	mean_bns, sigma_bns = bns
	mean_nsbh, sigma_nsbh = nsbh
	mean_bbh, sigma_bbh = bbh
	return  {"Terrestrial": noise_dist(),
		"BNS": signal_dist(mean_bns, sigma_bns),
		"NSBH": signal_dist(mean_nsbh, sigma_nsbh),
		"BBH": signal_dist(mean_bbh, sigma_bbh)}

def p_c(p_x_c, ranking_stat_thresh= 4, N_events={"Terrestrial": 1e6, "BNS": 10, "NSBH": 10, "BBH": 1000}):
	N_above_thresh = {"Terrestrial": N_events["Terrestrial"],
			"BNS": N_events["BNS"],
			"NSBH": N_events["NSBH"],
			"BBH": N_events["BBH"]}
	p_c = {k: N_above_thresh[k]/sum(N_above_thresh.values()) / (p_x_c[k][:,ranking_stat_thresh:]).sum() for k in p_x_c}
	return p_c

def p_astro(mc, L, p_x_c, p_c, mc_centers = mc_centers, L_centers = L_centers):
	mc_ix = numpy.searchsorted(mc_centers, mc)
	L_ix = numpy.searchsorted(L_centers, L)
	out = {k: p_x_c[k][mc_ix, L_ix] * p_c[k] for k in p_x_c}
	norm = sum(out.values())
	return {k: out[k]/norm for k in out}




#!/usr/bin/env python

import htcondor
import itertools
import yaml
from yaml.loader import SafeLoader

def add_job_opts(job):

	opts = {'executable': f'env/bin/{job}',
		'error':f'logs/{job}-$(cluster)-$(process).err',
		'output':f'logs/{job}-$(cluster)-$(process).out',
		}
	return opts

def add_job_args(opts):
	args = ''
	for (k, v) in opts.items():
		if isinstance(v, list):
			for vv in v:
				args += f'--{k.replace("_", "-")} {vv} '
		elif type(v) == bool and v:
			args += f'--{k.replace("_", "-")} '
		elif type(v) == bool:
			continue
		else:
			args += f'--{k.replace("_", "-")} {v} '
	return {'arguments': args}

def get_topics(datasource, topic_suffix, tag):
	topics = []
	if len(datasource) >= len(topic_suffix):
		combinations = [list(zip(p, topic_suffix)) for p in itertools.permutations(datasource, len(topic_suffix))]
	else:
		combinations = [list(zip(datasource, p)) for p in itertools.permutations(topic_suffix, len(datasource))]
	for c in combinations:
		for pipeline, topic in c:
			topics.append(f'{pipeline}.{tag}.{topic}')
	return topics

def add_common_args(config, job):
	job_args = {}
	job_args.update({
				'tag': config['tag'],
				'kafka_server': config['kafka_server'],
	})

	for arg, val in config['jobs'][job].items():
		# input topic arg is handled differently
		if arg == 'input_topic':
			continue
		job_args.update({arg: val})

	return job_args

def send_inj_stream_layer(dag, config, opts):
	# add job options
	opts.update(add_job_opts('send-inj-stream'))

	job_args = {}
	
	# add job arguments
	job_args.update(add_common_args(config, 'send_inj_stream'))
	job_args.update({'data_source': config['data-source']})
	job_args.update({'inj-file': config['injections']})
	job_args.update({'ifo': [ifo for ifo in config['ifos'].split(',')]})
	if 'track-psd' in config['jobs']['send_inj_stream'] or 'track-segments' in config['jobs']['send_inj_stream']:
		job_args.update({'input-topic': [topic for topic in config['jobs']['send_inj_stream']['input_topic']]})

	opts.update(add_job_args(job_args))

	# add job to the dag
	dag.layer(name='send_inj_stream', submit_description = htcondor.Submit(opts), retries = '3')
	return dag

def inspinjmsg_find_layer(dag, config, opts):
	# add job options
	opts.update(add_job_opts('inspinjmsg-find'))

	job_args = {}
	
	# add job arguments
	job_args.update(add_common_args(config, 'inspinjmsg_find'))
	tag = config['tag']
	topics = get_topics(config['data-source'], config['jobs']['inspinjmsg_find']['input_topic'], tag)
	job_args.update({'input_topic': topics})
	opts.update(add_job_args(job_args))

	# add job to the dag
	dag.layer(name='inspinjmsg_find', submit_description = htcondor.Submit(opts), retries = '3')
	return dag

def igwn_alert_listener_layer(dag, config, opts):
	# add job options
	opts.update(add_job_opts('igwn-alert-listener'))

	job_args = {}
	
	# add job arguments
	job_args.update(add_common_args(config, 'igwn_alert_listener'))
	job_args.update({'group': config['group']})
	opts.update(add_job_args(job_args))

	# add job to the dag
	dag.layer(name='igwn_alert_listener', submit_description = htcondor.Submit(opts), retries = '3')
	return dag

def inj_missed_found_layer(dag, config, opts):
	# add job options
	opts.update(add_job_opts('inj-missed-found'))

	job_args = {}
	
	# add job arguments
	job_args.update(add_common_args(config, 'inj_missed_found'))
	tag = config['tag']
	topics = get_topics(config['data-source'], config['jobs']['inj_missed_found']['input_topic'], f'{tag}.testsuite')
	job_args.update({'input_topic': topics})
	opts.update(add_job_args(job_args))

	# add job to the dag
	dag.layer(name='inj_missed_found', submit_description = htcondor.Submit(opts), retries = '3')
	return dag

def vt_layer(dag, config, opts):
	# add job options
	opts.update(add_job_opts('vt'))

	job_args = {}
	
	# add job arguments
	job_args.update(add_common_args(config, 'vt'))
	tag = config['tag']
	topics = get_topics(config['data-source'], config['jobs']['vt']['input_topic'], f'{tag}.testsuite')
	job_args.update({'input_topic': topics})
	job_args.update({'inj-file': config['injections']})
	opts.update(add_job_args(job_args))

	# add job to the dag
	dag.layer(name='vt', submit_description = htcondor.Submit(opts), retries = '3')
	return dag

def latency_layer(dag, config, opts):
	# add job options
	opts.update(add_job_opts('latency'))

	job_args = {}
	
	# add job arguments
	job_args.update(add_common_args(config, 'latency'))
	tag = config['tag']
	topics = get_topics(config['data-source'], config['jobs']['latency']['input_topic'], f'{tag}.testsuite')
	job_args.update({'input_topic': topics})
	opts.update(add_job_args(job_args))

	# add job to the dag
	dag.layer(name='latency', submit_description = htcondor.Submit(opts), retries = '3')
	return dag

def likelihood_layer(dag, config, opts):
	# add job options
	opts.update(add_job_opts('likelihood'))

	job_args = {}
	
	# add job arguments
	job_args.update(add_common_args(config, 'likelihood'))
	tag = config['tag']
	topics = get_topics(config['data-source'], config['jobs']['likelihood']['input_topic'], f'{tag}.testsuite')
	job_args.update({'input_topic': topics})
	opts.update(add_job_args(job_args))

	# add job to the dag
	dag.layer(name='likelihood', submit_description = htcondor.Submit(opts), retries = '3')
	return dag

def p_astro_layer(dag, config, opts):
	# add job options
	opts.update(add_job_opts('p-astro'))

	job_args = {}
	
	# add job arguments
	job_args.update(add_common_args(config, 'p_astro'))
	tag = config['tag']
	topics = get_topics(config['data-source'], config['jobs']['p_astro']['input_topic'], f'{tag}.testsuite')
	job_args.update({'input_topic': topics})
	opts.update(add_job_args(job_args))

	# add job to the dag
	dag.layer(name='p_astro', submit_description = htcondor.Submit(opts), retries = '3')
	return dag

def skymap_layer(dag, config, opts):
	# add job options
	opts.update(add_job_opts('skymap'))

	job_args = {}
	
	# add job arguments
	job_args.update(add_common_args(config, 'skymap'))
	tag = config['tag']
	topics = get_topics(config['data-source'], config['jobs']['skymap']['input_topic'], f'{tag}.testsuite')
	job_args.update({'input_topic': topics})
	if 'gdb-skymaps' in job_args.keys():
		job_args.update({'group': config['group']})
	opts.update(add_job_args(job_args))

	# add job to the dag
	dag.layer(name='skymap', submit_description = htcondor.Submit(opts), retries = '3')
	return dag

def snr_consistency_layer(dag, config, opts):
	for i, ifo in enumerate(config['ifos'].split(',')):
		# add job options
		opts.update(add_job_opts('snr-consistency'))
	
		job_args = {}

		# add job arguments
		job_args.update(add_common_args(config, 'snr_consistency'))
		tag = config['tag']
		topics = get_topics(config['data-source'], [f'{ifo}_{t}' for t in config['jobs']['snr_consistency']['input_topic']], f'{tag}.testsuite')

		job_args.update({'input_topic': topics})
		job_args.update({'ifo': ifo})
		opts.update(add_job_args(job_args))

		# add job to the dag
		dag.layer(name=f'snr_consistency_{int(i):04d}', submit_description = htcondor.Submit(opts), retries = '3')

	return dag

def inj_accuracy_layer(dag, config, opts):
	for i, ifo in enumerate(config['ifos'].split(',')):
		# add job options
		opts.update(add_job_opts('inj-accuracy'))

		job_args = {}

		# add job arguments
		job_args.update(add_common_args(config, 'inj_accuracy'))
		tag = config['tag']
		topics = get_topics(config['data-source'], config['jobs']['inj_accuracy']['input_topic'], f'{tag}.testsuite')
		job_args.update({'input_topic': topics})
		job_args.update({'ifo': ifo})
		opts.update(add_job_args(job_args))

		# add job to the dag
		dag.layer(name=f'inj_accuracy_{int(i):04d}', submit_description = htcondor.Submit(opts), retries = '3')

	return dag

def collect_metrics_layer(dag, config, config_path, opts):
	# load options from the config

	kafka_server = config['kafka_server']
	tag = config['tag']

	for i, source in enumerate(config['data-source']):
		web_config_path = config['metrics'][source]['config']
		# grab the web config
		with open(web_config_path, 'r') as f:
			web_config = yaml.load(f, Loader=SafeLoader)

		# add job options
		opts.update({'executable': 'env/bin/scald',
			'error':'logs/scald_metric_collector-$(cluster)-$(process).err',
			'output':'logs/scald_metric_collector-$(cluster)-$(process).out',
			})
		arguments = f'aggregate --config {web_config_path} --uri kafka://{tag}@{kafka_server} --data-type timeseries'

		metrics = [metric for metric in web_config['schemas']]
		metrics.remove('inj_table_events')
		for metric in metrics:
			arguments += f' --topic {source}.{tag}.testsuite.{metric} --schema {metric}'

		opts.update(opts)
		opts.update({'arguments': arguments})

		dag.layer(name=f'scald_metric_collector_{i:04d}', submit_description = htcondor.Submit(opts), retries = '3')

	return dag

def scald_inj_table_events_layer(dag, config, config_path, opts):
        #load options from config
        kafka_server = config['kafka_server']
        tag = config['tag']

        for i, source in enumerate(config['data-source']):
                web_config_path = config['metrics'][source]['config']
                # grab the web config
                with open(web_config_path, 'r') as f:
                      web_config = yaml.load(f, Loader=SafeLoader)

                # add job options
                opts.update({'executable': 'env/bin/scald',
                     'error':'logs/scald_event_collector-$(cluster)-$(process).err',
                     'output':'logs/scald_event_collector-$(cluster)-$(process).out',
                     })
                arguments = f'aggregate --config {web_config_path} --uri kafka://{tag}@{kafka_server} --data-type triggers'

                arguments += f' --topic {source}.{tag}.testsuite.inj_table_events --schema inj_table_events'

                opts.update(opts)
                opts.update({'arguments': arguments})

                dag.layer(name=f'scald_event_collector_{i:04d}', submit_description = htcondor.Submit(opts), retries = '3')

        return dag






#!/usr/bin/env python3

from optparse import OptionParser
import os
import sys
import json
import numpy
import copy
import logging

from time import sleep
from collections import defaultdict, deque
from confluent_kafka import Producer

from cronut import App
from cronut.utils import uriparse

from ligo.lw import lsctables
from lal import GPSTimeNow
from lal import LIGOTimeGPS

from ligo.scald.io import kafka

from gw.lts import utils

def parse_command_line():
	parser = utils.add_general_opts()
	parser.add_option('--ifo', help = 'Interferometer to get data from')
	opts, args = parser.parse_args()

	return opts, args

class SNRConsistency(object):
	def __init__(self, options):
		self.ifo = options.ifo
		self.tag = options.tag
		self.kafka_server = options.kafka_server
		self.topics = options.input_topic

		# set up producer
		self.client = kafka.Client(f'kafka://{self.tag}@{self.kafka_server}')

		# initialize data deque and output dict
		self.data = defaultdict(lambda: defaultdict(lambda: deque(maxlen=300)))
		
		# create a job service using cronut
		self.app = App('snr_consistency', broker=f'kafka://{self.tag}_{self.ifo}_snr_consistency@{self.kafka_server}')

		# subscribes to a topic
		@self.app.process(self.topics)
		def process(message):
			mifo, mtopic = message.topic().split('.')[-1].split('_')
			mpipeline = message.topic().split('.')[0]
			mkey = message.key().decode('utf-8')
		
			source, is_recovered = mkey.split('.')
			logging.info(f'Read message from input {mpipeline} {mtopic}: {source} {is_recovered}')

			# unpack data from the message 
			# store SNRs in a dictionary keyed by event time
			m = json.loads(message.value())
			val = m['data'][-1]
			time = m['time'][-1]

			if is_recovered == 'missed':
				logging.debug(f"Injection at {time} missed, skipping.")
			else:
				time_window = utils.event_window(time)

				self.data[mpipeline][mtopic].append({
								'time': time,
								'snr': val
				})

				self.process_msgs(mpipeline)


	def start(self):
		# start up
		logging.info('Starting up...')
		self.app.start()


	def process_msgs(self, pipeline):
		# for each reecovered msg time in the deque
		# find the nearest injection in injected msgs deque
		# within +/- delta_t (1 second) of the recovered msg
		# time. When an association is made, remove the recovered
		# msg from the deque, calculate the snr accuracy and
		# send a message to the output topic
		recovered = copy.copy(self.data[pipeline]['recsnr'])
		for rec in recovered:

			rec_time = rec['time']
			rec_snr = rec['snr']

			nearest_inj = utils.find_nearest_msg(self.data[pipeline]['injsnr'], rec_time)

			if not nearest_inj:
				continue

			inj_time = nearest_inj['time']
			inj_snr = nearest_inj['snr']

			self.data[pipeline]['recsnr'].remove(rec)

			accuracy = 1. - (inj_snr - rec_snr) / inj_snr

			output = {
					'time': [ inj_time ],
					'data': [ accuracy ]
			}

			self.client.write(f'{pipeline}.{self.tag}.testsuite.{self.ifo}_snr_accuracy', output)
			logging.info(f'Sent msg to: {pipeline}.{self.tag}.testsuite.{self.ifo}_snr_accuracy')

def main():
	opts, args = parse_command_line()

	# sanity check input options
	required_opts = ['ifo', 'tag', 'input_topic', 'kafka_server']
	for r in required_opts:
		if not getattr(opts, r):
			raise ValueError(f'Missing option: {r}.')

	# set up logging
	utils.set_up_logger(opts.verbose)

	# start up processor
	processor = SNRConsistency(opts)
	processor.start()

if __name__ == '__main__':
	main()
